/*
  **************************************************************
  *** Main program for 2D Hessian application                ***
  *** It is employed during linearized waveform inversion    ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2018 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "lint2.h"
#include "interp_psf.h"
#include "hess_op.h"
#include <sep.main>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {
  // Define class member objects
  Read par;
  Write out;

  std::cout << "*************************************************************" << std::endl;
  std::cout << "*** 2D application of the Gauss-Newton Hessian              *" << std::endl;
  std::cout << "*** Utilized during Linearized Waveform Inversion           *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                              *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2018  *" << std::endl;
  std::cout << "*************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Build geomery from input parameters
  par.build_geometry();

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize model and data vectors
  int nm=par.nxE*par.nzE*(2*par.nh+1);
  int nd=(par.nxE+2*par.nlagx)*(par.nzE+2*par.nlagz)*(2*par.nh+1);
  std::vector<float> model(nm,0.);
  std::vector<float> data (nd,0.);

  // Read reflectivity (actually reads the RTM image)
  par.read_refl_hess(nd);

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;

  // Initialize PSF interpolator
  InterpPsf psf(par,1);

  // Initialize Hess operator
  Hess_op op(par,psf,1); 

  if(par.adj==1) {
  // Adjoint Hessian
    if(nd > par.refl.size()) {
      for(int jh=0; jh<2*par.nh+1; jh++) {
        for(int jx=par.nlagx; jx<par.nxE+par.nlagx; jx++) {
          for(int jz=par.nlagz; jz<par.nzE+par.nlagz; jz++) {
            int j1=jz-par.nlagz+(jx-par.nlagx)*par.nzE+jh*par.nzE*par.nxE;
            int j2=jz+jx*(par.nzE+2*par.nlagz)+
                +jh*(par.nzE+2*par.nlagz)*(par.nxE+2*par.nlagx);
            data[j2]=par.refl[j1];
          }
        }
      }
    }
    else if(nd == par.refl.size()) {
      data.swap(par.refl);
    }
    else {
      seperr("Hessian2d: There's an error in the input reflectivity dimensions");
    }
    op.op(1,0,model,data);
  }
  else {
  // Forward Hessian
    model.swap(par.refl);
    op.op(0,0,model,data);
  } 

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;
  // Initialize the output
  out.write_init(par);
  
  // Write the output
  if(par.adj==1) {
    std::cout << "Write the reflectivity" << std::endl;  
    out.write_hess_image(par,model.data(),1);
  }
  else {
    std::cout << "Write the reflectivity" << std::endl;  
    out.write_hess_image(par,data.data(),0);
  }

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "Hessian2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}

