/*
  ***************************************************************
  *** Main program for the dot-product test for rtm and wemva ***
  *** Version rewritten in C++                                ***
  *** Author: Alejandro Cabrales                              ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017  ***
  ***************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "interp.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "lap.h"
#include "grad.h"
#include "grad2.h"
#include "wave_prop.h"
#include "scatter.h"
#include "scale.h"
#include "vshot.h"
#include "rshot.h"
#include "alloc.h"
#include <cmath>
#include <vector>
#include "lint2.h"
#include "nlmod.h"
#include "rtm.h"
#include "born.h"
#include "wemva.h"
#include "born_op.h"
#include "dso.h"
#include "wemva_op.h"
#include "hess_op.h"
#include "lwivu_op.h"
#include "matmult.h"
#include "operator.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <numeric>

int MAIN() {
  // Define class member objects
  int op;
  float rand1;
  Read par;
  VShot vel;
  WaveProp fwd, scat, bwd;
  Born born;
  Rtm rtm; 
  Wemva wemva,wemva_adj;
  Write out;
  Grad grad;
  NLMod nlmod; 

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();
  if(0 == getch("op","d",&op)) 
    op=2; 

  // Read data history
  par.read_history();

  // Read data header
  par.read_headers();
  
  // Read velocity
  par.read_velocity();
 
  // Read reflectivity
  if(op == 1 || op == 2 || op == 3 || op == 4) {
    par.read_reflectivity("none");
  }

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Obtain data geometry and apertures from input parameters
  par.data_geometry();
  
  // Read and verbose source
  par.read_src(1);
  par.jtStep=1;

  // Verbose some useful parameters
  par.verbose_par();
 
  // Initialize tapering for receiver propagation
//  Read &ppar=par; // Reference to make changes in "par" members within a function 
  par.use_taper=0;
//  taper_init(ppar);

  if(op != 7 && op != 8) {
    // Initialize variables for shot velocities
    vel.vshot_init(par);
  
    // Initialize parameters for source and scattered wavefields
    fwd.prop_init(par);
    scat.prop_init(par);
    bwd.prop_init(par);
  
    // Create velocity frames for source wavefield
    vel.vshot(par,par.Vint,fwd.Vshot,par.sgx,0); // Don't use random boundary cond.
  
    // Create velocity frames for source wavefield
    vel.vshot(par,par.Vint,bwd.Vshot,par.sgx,0); // Don't use random boundary cond.
  
    // Create velocity frames for scattered source wavefield
    vel.vshot(par,par.Vint,scat.Vshot,par.sgx,0); // Don't use random boundary cond.
  }

  Grad2 grad2(par);

  // Alocate arrays for the dot-product test
  std::vector<float> r, rm, d, dm, de, dem, B;
  int n1=par.nkgx*(2*par.nh+1)*par.nkz;
  int n2=par.nkgx*par.nt;
  int n3=10000;

  if(op==1 || op==2 || op==3 || op==5) {
    allocV(r ,n1);
    allocV(rm,n1);
  }
  else if(op==4) {
    allocV(r ,2*n1);
    allocV(rm,2*n1);
  }
  else if(op==7) {
    allocV(r ,par.ntData);
    allocV(rm,par.ntData);
  }
  else if(op==8) {
    allocV(r ,n3);
    allocV(rm,n3);
  }
  else {
    allocV(r ,par.nt);
    allocV(rm,par.nt);
  }

  switch(op) {
    case 1:  // Wemva
      std::cout << "Allocate arrays for WEMVA" << std::endl;
      allocV(de ,n1);
      allocV(dem,n1);
      break; 
    case 2: // Born
      std::cout << "Allocate arrays for Born" << std::endl;
      allocV(d ,n2);
      allocV(dm,n2);
      break; 
    case 3: // Hessian
      std::cout << "Allocate arrays for Hessian" << std::endl;
      allocV(de, n1);
      allocV(dem,n1);
      break; 
    case 4: // LWIVU
      std::cout << "Allocate arrays for LWIVU" << std::endl;
      allocV(de, 2*n1);
      allocV(dem,2*n1);
      break; 
    case 5:
      std::cout << "Allocate arrays for Gradient" << std::endl;
      allocV(de ,n1);
      allocV(dem,n1);
      break; 
    case 6:
      std::cout << "Allocate arrays for nlmodel" << std::endl;
      allocV(d ,n2);
      allocV(dm,n2);
      break; 
    case 7:
      std::cout << "Allocate arrays for interp" << std::endl;
      allocV(d ,par.nt);
      allocV(dm,par.nt);
      break; 
    default:
      std::cout << "Allocate arrays for matmult" << std::endl;
      allocV(d ,n3);
      allocV(dm,n3);
      allocV(B,n3*n3);
      break; 
  }

  // Produce random data for the test
  std::cout << "Produce random data for the test" << std::endl;
  srand(time(NULL));

  for(int jj=0; jj<rm.size(); jj++) {
    rand1=((float) rand()/(RAND_MAX));
    rand1=2.0*(rand1-0.5);
    rm[jj]=rand1; 
  }

  for(int jj=0; jj<dm.size(); jj++) {
    rand1=((float) rand()/(RAND_MAX));
    rand1=2.0*(rand1-0.5);
    dm[jj]=rand1; 
  }

  for(int jj=0; jj<dem.size(); jj++) {
    rand1=((float) rand()/(RAND_MAX));
    rand1=2.0*(rand1-0.5);
    dem[jj]=rand1; 
  }

  // Initialize some variables for the main calculation
  par.ntSnap=0;
  par.first_geof=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Get number of geophone traces in shot gather (in some datasets is variable)
  par.ngeof = par.source_index[0];
  par.last_geof = par.first_geof+par.ngeof;

  int doit=0;
  if(op == 3 || op == 4) doit=1;
  Hess_op hess(par,doit);
  doit=0;
  if(op == 4) doit=1;
  Dso dso(par,1);
  Wemva_op wemva_op(par,wemva,wemva_adj,fwd,bwd,scat,dso,doit); 
  Lwivu_op lwivu(par,hess,wemva_op,dso,doit);
  doit=0;
  if(op == 7) doit=1;
  Interp interp_op(par.tIntFact,doit);
  doit=0;
  if(op == 8) {
    doit=1;
    for(int jj=0; jj<B.size(); jj++) {
      rand1=((float) rand()/(RAND_MAX));
      rand1=2.0*(rand1-0.5);
      B[jj]=rand1; 
    }
  }   
  Matmult matmult_op(B,doit);

     std::vector<float> ptest1(par.nkgx*par.nkz,0.);
     std::vector<float> ptest2(par.nkgx*par.nkz,0.);
  switch(op) {
    case 1: // WEMVA
    {
      // Initialize Wemva operator
      Read &ppar=par;
      Wemva &pwemva=wemva;
      Wemva &pwemva_adj=wemva_adj;
      Dso &pdso=dso;
      Wemva_op op(ppar,pwemva,pwemva_adj,fwd,bwd,scat,dso,1); 
    
      // Pointer to the operator
      void (Wemva_op::*wemva_op)(int,int,
                     std::vector<float>&,
                     std::vector<float>&)=&Wemva_op::op;
    
      // Forward modeling
      op.op(0,0,rm,de);
    
      // Adjoint modeling
      op.op(1,0,r,dem);
    break;
    }
    case 2: // Born
    {
      // Initialize Born operator
      Read &ppar=par;
      Born &pborn=born;
      Rtm &prtm=rtm;
      Born_op op(ppar,prtm,pborn,fwd,bwd,scat); 
    
      // Pointer to the operator
      void (Born_op::*born_op)(int,int,
                     std::vector<float>&,
                     std::vector<float>&)=&Born_op::op;
    
      // Forward modeling
      op.op(0,0,rm,d);
    
      // Adjoint modeling
      op.op(1,0,r,dm);
    break; 
    } 
    case 3: // Hessian mult
      // Initialize parameters for Hessian
      std::cout << "Initialize parameters for Hessian" << std::endl;

      // Forward modeling
      std::cout << "Run forward Hessian" << std::endl;
      hess.op(0,0,rm,de);

      // Adjoint modeling
      std::cout << "Run adjoint Hessian" << std::endl;
      hess.op(1,0,r,dem);
    break;  
    case 4: // LWIVU
      // Initialize parameters for LWIVU
      std::cout << "Initialize parameters for LWIVU" << std::endl;

      // Forward modeling
      std::cout << "Run forward LWIVU" << std::endl;
      lwivu.op(0,0,rm,de);

      // Adjoint modeling
      std::cout << "Run adjoint LWIVU" << std::endl;
      lwivu.op(1,0,r,dem);
    break;  
    case 5:
      std::cout << "Initialize parameters for Gradient" << std::endl;

      grad.grad_init(par);
      // Forward modeling
      ptest1=rm;
//      grad.gradz(ptest1,ptest2,0);
      grad2.grad2(ptest1,ptest2);
        de=ptest2;

      // Adjoint modeling
      ptest1=dem;
//      grad.gradz(ptest1,ptest2,1);
      grad2.grad2(ptest1,ptest2);
      r=ptest2;
      break; 
    case 6:
      std::cout << "Initialize parameters for Non-linear modeling" << std::endl;
      nlmod.nlmod_init(par);

      // Forward modeling
      par.src=rm;
      par.adj=0;
      nlmod.nlmod(par,0);
      d=nlmod.dsynt;

      // Adjoint modeling
      par.rcv=dm;
      par.adj=1;
      nlmod.nlmod(par,0);
      r=par.src;
    break;
    case 7:
      std::cout << "Initialize parameters for interp" << std::endl;

      // Forward modeling
      std::cout << "Run forward interpolation" << std::endl;
      interp_op.op(0,0,rm,d);

      // Adjoint modeling
      std::cout << "Run adjoint interpolation" << std::endl;
      interp_op.op(1,0,r,dm);
    break;
    default:
      std::cout << "Initialize parameters for matmult" << std::endl;

      // Forward modeling
      std::cout << "Run forward matrix multip." << std::endl;
      matmult_op.op(0,0,rm,d);

      // Adjoint modeling
      std::cout << "Run adjoint matrix multip." << std::endl;
      matmult_op.op(1,0,r,dm);
    break;
  } 

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;
  // Convert results into one-dimansional vectors and perform dot products
  float dpData;
  if(op==1 || op==4 || op==3) {
    dpData=std::inner_product(de.begin(),de.end(),dem.begin(),0.0);
  }
  else { // if(op==2 || op==5) {
    dpData=std::inner_product(d.begin(),d.end(),dm.begin(),0.0);
  }
  
  float dpRefl=std::inner_product(r.begin(),r.end(),rm.begin(),0.0);

  std::cout << "rdr: "   << std::inner_product(r.begin(), r.end(), r.begin(), 0.0) << std::endl;
  std::cout << "rmdrm: " << std::inner_product(rm.begin(),rm.end(),rm.begin(),0.0) << std::endl;
  
  // Obtain the absoluteand relative errors
  float err_abs=dpData-dpRefl; 
  float err_rel=err_abs/dpData;

//  // Report the results
  std::cout << "dpData: " << dpData << std::endl;
  std::cout << "dpRefl: " << dpRefl << std::endl;
  std::cout << "err_abs: " << err_abs << std::endl;
  std::cout << "err_rel: " << err_rel << std::endl;

  // Write snapshots
//  out.write_snap("snap1");
//  out.write_snap("snap2");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "DPtest2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}

