/*
  **************************************************************
  *** Main program for 2D Hessian-WEMVA application          ***
  *** It is employed during linearized waveform inversion    ***
  *** with velocity updating                                 ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2018 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "interp1d.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include <cmath>
#include "snap.h"
#include "alloc.h"
#include "allocv.h"
#include "lint2.h"
#include "wemva.h"
#include "operator.h"
#include "dso.h"
#include "wemva_op.h"
#include "lwivu_op.h"
#include <vector>
#include "map.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <functional>
#include <numeric>

int MAIN() {
  // Define class member objects
  Read par;
  WaveProp fwd, scat, bwd;
  Wemva wemva,wemva_adj;
  Write out;

  std::cout << "*****************************************************************" << std::endl;
  std::cout << "*** 2D application of the Gauss-Newton Hessian and WEMVA        *" << std::endl;
  std::cout << "*** Utilized during Linearized Waveform Inv. with Vel. Updating *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                  *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2018      *" << std::endl;
  std::cout << "*****************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Build geomery from input parameters
  par.build_geometry();

  // Read velocity
  par.read_velocity();

  // Read reflectivity (actually reads the RTM image)
  par.read_reflectivity("reflfile");

  // Read perturbation (either in slowness or reflectivity)
  par.read_pert(par.adj);
  
  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Read and verbose source
  par.read_src(1);
  
  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize tapering for receiver propagation
  taper_init(par);

  // Initialize some variables for the main loop
  par.ntSnap=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Initialize parameters for forward and backward propagations
  fwd.prop_init(par);
  bwd.prop_init(par);
  scat.prop_init(par);

  // Create velocity frames
  par.vel_frames(fwd.Vshot,par.rbc);
  par.vel_frames(bwd.Vshot,par.rbc);
  par.vel_frames(scat.Vshot,0);

  // Initialize model and data vectors
  int n=par.nxE*(2*par.nh+1)*par.nzE;
  std::vector<float> model(2*n,0.);
  std::vector<float> data;

  // Initialize LWIVU operator
  Hess_op hess_op(par,1); 
  Dso dso(par,0);
  Wemva_op wemva_op(par,wemva,wemva_adj,fwd,bwd,scat,dso,1); 
  Lwivu_op op(par,hess_op,wemva_op,dso,1);

  // Create image frame (data1)
  std::vector<float> temp1(n,0.);
  std::vector<float> temp2(n,0.);
  par.refl_frames(temp1);
  temp2=temp1;

  // Compute data2 (multiply image by -1) 
  std::transform(temp2.begin(),temp2.end(),temp2.begin(),
  std::bind1st(std::multiplies<float>(),-1.));

  // Insert into temp vector
  temp1.insert(temp1.end(),temp2.begin(),temp2.end());
  temp2.clear();

  if(par.adj==1) {
  // Adjoint LWIVU
    data.swap(temp1);
    temp1.clear();
    op.op(1,0,model,data);
  }
  else {
  // Forward LWIVU
    model.swap(temp1);
    temp1.clear();
    op.op(0,0,model,data);
  } 

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;
  // Initialize the output
  out.write_init(par);
  
  // Write the output
  if(par.adj==1) {
    std::cout << "Write the reflectivity" << std::endl;  
    out.write_image(par,model.data(),"hessian");
  }
  else {
    std::cout << "Write the reflectivity" << std::endl;  
    out.write_image(par,data.data(),"hessian");
  }

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "Hessian2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}

