/*
  ***************************************************************
  *** Main program for the linear test for rtm and wemva      ***
  *** Version rewritten in C++                                ***
  *** Author: Alejandro Cabrales                              ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017  ***
  ***************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "lap.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "vshot.h"
#include "rshot.h"
#include "alloc.h"
#include <cmath>
#include "nlmod.h"
#include "rtm.h"
#include "born.h"
#include "wemva.h"
#include "born_op.h"
#include "dso.h"
#include "wemva_op.h"
#include "operator.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <numeric>

int MAIN() {
  // Define class member objects
  int op;
  Read par;
  VShot vel;
  WaveProp fwd, scat, bwd;
  NLMod nlmod; 
  Born born;
  Rtm rtm; 
  Wemva wemva;
  Write out;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();
  if(0 == getch("op","d",&op)) op=2;

  // Read data history
  par.read_history();

  // Read data header
  par.read_headers();
  
  // Read velocity
  par.read_velocity();

  // Set a constant velocity model
  for(int jx=0; jx<par.nx; jx++) {
    for(int jz=0; jz<par.nz; jz++) {
      par.Vint[jx][jz] = 2500.0;
    }
  }
  par.Vcte=2500.0;
  par.Vmin=2500.0;
  par.Vmax=2500.0;
 
  int n1=par.nz/2;
  int n2=par.nx/2;

  // Determine time sampling according to Courant-Fredrich-Lewis condition
//  par.Vint[n2][n1] = par.Vcte*1.03;
  par.courant();
//  par.Vint[n2][n1] = par.Vcte;
  
  // Read reflectivity
  par.read_reflectivity("none");

  // Obtain data geometry and apertures from input parameters
  par.data_geometry();
  
  float V, Vpdv, dS2;
  std::vector<float> temp1(par.nx*(2*par.nh+1)*par.nz,0.);
  std::vector<float> temp2(par.nkgx*par.nt,0.);

  // Create the corresponding slowness squared field
  for(int jx=0; jx<par.nx; jx++) {
    for(int jh=0; jh<(2*par.nh+1); jh++) {
      for(int jz=0; jz<par.nz; jz++) {
        par.refl[jx][jh][jz] = 0.0;
      }
    }
  }

  float pert=0.03;
  Vpdv=(1.+pert)*par.Vcte;
  V=par.Vcte;
  float delS2=(1.0/Vpdv/Vpdv-1.0/V/V);
  par.refl[n2][par.nh][n1]=delS2;

  // Verbose some useful parameters
  par.verbose_par();
 
  // Initialize tapering for receiver propagation
  par.use_taper=0;
//  taper_init(par);

  // Initialize some variables for the main loop
  par.ntSnap=0;
  par.first_geof=0;

  // Initialize variables for shot velocities
  vel.vshot_init(par);

  // Initialize parameters for source and scattered wavefields
  fwd.prop_init(par);
  scat.prop_init(par);
  bwd.prop_init(par);

  // Create velocity frames for source wavefield
  vel.vshot(par,par.Vint,fwd.Vshot,par.sgx,0); // Don't use random boundary cond.

  // Create velocity frames for source wavefield
  vel.vshot(par,par.Vint,bwd.Vshot,par.sgx,0); // Don't use random boundary cond.

  // Create velocity frames for scattered source wavefield
  vel.vshot(par,par.Vint,scat.Vshot,par.sgx,0); // Don't use random boundary cond.

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Get number of geophone traces in shot gather (in some datasets is variable)
  par.ngeof = par.source_index[0];
  par.last_geof = par.first_geof+par.ngeof;

  std::vector<float> dexact, dapprox;

  // Compute approximate operators
  switch(op) {
    case(1):
      // Prepare arrays for the output
      allocV(dexact,par.nx*(2*par.nh+1)*par.nz); 
      allocV(dapprox,par.nx*(2*par.nh+1)*par.nz); 

      // Read source and apply 2nd derivative
      par.read_src(1);
      par.jtStep=1;

      // Initialize parameters for wemva
      wemva.wemva_init(par,fwd,bwd,scat);

      // Perform WEMVA modeling
      wemva.wemva(par);

      // Build wemva perturbation
      wemva.build_pert(par);

      // Save perturbation as approximated
      dapprox=wemva.IPert;
    break;  
    
    case(2):
      // Prepare arrays for the output
      allocV(dexact,par.nkgx*par.nt); 
      allocV(dapprox,par.nkgx*par.nt); 

      // Read source, don't apply 2nd derivative
      par.read_src(1);
      par.jtStep=1;

      // Initialize parameters for born modeling
      born.born_init(par,fwd,scat);

      // Perform Born modeling
      born.born(par);
      for(int jj=0; jj<dapprox.size(); jj++) {
        dapprox[jj]=born.dsynt[jj];
      } 
    break;  
  } 

  // Perform non-linear modeling on unperturbed model
  switch(op) {
    case(1):
      // Read source, apply 2nd derivative
      par.read_src(1);
      par.jtStep=1;

      // Initialize parameters for rtm
      rtm.rtm_init(par,fwd,bwd);

      // Perform Reverse-Time Migration in each shot gather
      rtm.rtm(par);

      // Construct the image
      rtm.build_images(par);

      // Temporally save rtm image
//      temp1=rtm.ICorr;
      temp1.swap(rtm.ICorr);
    break;  
    
    case(2):
      // Read source, don't apply 2nd derivative
      par.read_src(0);

      // Initialize parameters for non-linear modeling
      nlmod.nlmod_init(par);

      // Perform non-linear modeling
      nlmod.nlmod(par,0);

      for(int jj=0; jj<temp2.size(); jj++) {
        temp2[jj]=nlmod.dsynt[jj];
      }
    break;  
  } 

  // Include the perturbation of the velocity
  par.Vint[n2][n1] = par.Vcte*(1.+pert);

  // Perform forward modeling on perturbed model
  switch(op) {
    case(1):
      // Re-initialize parameters for rtm
      rtm.rtm_init(par,fwd,bwd);

      // Perform Reverse-Time Migration in each shot gather
      rtm.rtm(par);

      // Construct the image
      rtm.build_images(par);

      // Obtain the difference with the unperturbed image
      for(int jj=0; jj<temp1.size(); jj++) {
        rtm.ICorr[jj]-=temp1[jj];
      } 
      dexact=rtm.ICorr;
    break;  
    
    case(2):
      // Re-initialize parameters for non-linear modeling
      nlmod.nlmod_init(par);

      // Perform non-linear modeling
      nlmod.nlmod(par,0);

      for(int jj=0; jj<temp2.size(); jj++) {
        nlmod.dsynt[jj]-=temp2[jj];
      } 
      dexact=nlmod.dsynt;
    break;  
  } 

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;

  // Initialize the output
  out.write_init(par);
  float **D;
   
  if(op == 1) {
  // Write the images history files and binaries for RTM/WEMVA
    out.write_image(par,rtm.ICorr.data(),"rtm_cic");
    out.write_image(par,wemva.IPert.data(),"wemva");
  }
  else if(op == 2) {
  // Write the images history files and binaries for RTM/WEMVA
    alloc(D,par.nkgx,par.nt);
    for(int jgx=0; jgx<par.nkgx; jgx++) {
      for(int jt=0; jt<par.nt; jt++) {
        int jj=jt+jgx*par.nt;
        D[jgx][jt] = nlmod.dsynt[jj];
      }
    }
    out.write_data(par,D,"nlmod");
    for(int jgx=0; jgx<par.nkgx; jgx++) {
      for(int jt=0; jt<par.nt; jt++) {
        int jj=jt+jgx*par.nt;
        D[jgx][jt] = born.dsynt[jj];
      }
    }
    out.write_data(par,D,"born");
  }

  // Perform linear test
  std::cout << "Perform the linear test" << std::endl;

  int n;
  float err_abs, err_rel;
  std::vector<float> de1, de2, diff;
  if(op == 1) {
    n=temp1.size();
  }
  else if(op == 2) {
    n=temp2.size();
  }
  allocV(diff,n);
  for(int jj=0; jj<n; jj++) {
    diff[jj]=dexact[jj]-dapprox[jj];
  } 

  // Obtain the absolute and relative errors
  err_abs=std::inner_product(diff.begin(),diff.end(),diff.begin(),0.0);
  err_rel=err_abs/std::inner_product(dexact.begin(),dexact.end(),dexact.begin(),0.0);

  float ddexact, ddapprox;
  ddexact=std::inner_product(dexact.begin(),dexact.end(),dexact.begin(),0.0);
  ddapprox=std::inner_product(dapprox.begin(),dapprox.end(),dapprox.begin(),0.0);

  // Report the results
  std::cout << "ddexact: " << ddexact << std::endl;
  std::cout << "ddapprox: " << ddapprox << std::endl;
  std::cout << "err_abs: " << err_abs << std::endl;
  std::cout << "err_rel: " << err_rel << std::endl;

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " 
            << std::chrono::duration<double>(t2-t1).count() 
            << "secs." << std::endl; 
  std::cout << "Lntest2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}
