/*
  **************************************************************
  *** Main program for 2D least-squares nonlinear modeling.  ***
  *** Using finite differences with Lax-Wendroff method      ***
  *** The objective is estimating a source from seismic data ***
  *** Version written in C++                                 ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <vector>
#include "lap.h"
#include "wave_prop.h"
#include "vshot.h"
#include <cmath>
#include "alloc.h"
#include "allocv.h"
#include "nlmod.h"
#include "operator.h"
#include "nlmod_op.h"
#include "tconv.h"
#include "solver_wjreg.h"
#include "cgstep.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {
  // Define class member objects
  Read par;
  NLMod nlmod; 
  Write out;
  Solver solver; 
  Conv tconv; 
  Step cgstep;
 
  std::cout << "**************************************************************" << std::endl;
  std::cout << "*** 2D least-squares nonlinear modeling for source estimation*" << std::endl;
  std::cout << "*** using FD with Lax-Wendroff method                        *" << std::endl;
  std::cout << "*** Version   written in C++                                 *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                               *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017   *" << std::endl;
  std::cout << "**************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
//  if(par.ntracesData==0) {// Read from parameters file
//     par.build_geometry();
//  }
//  else {                   // Read data header
    par.read_headers();
//  }

  // Read velocity
  par.read_velocity();
 
  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();

  // Proceed with the following if dataset is provided
//  if(par.ntracesData != 0) {
    par.data_geometry();

//    if(sep_copy_hff("datafile","nldata"))
//      seperr("nonlin2d: Cannot copy headers from datafile to nldatafile");
//  }

  // Read and verbose source (initializes necesary arrays; source will be overwritten)
  par.read_src(0);

  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize tapering
  Read &ppar=par; // Reference to make changes in "par" members within a function 
  taper_init(ppar);

  // Initialize parameters for non-linear modeling
  nlmod.nlmod_init(par);

  // Initialize some variables for the main loop
  par.ntSnap=0;
  par.first_geof=0;

  // Allocate auxiliary arrays
  std::vector<float> d(par.nt*par.nkgx);

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Get number of geophone traces in shot gather (in some datasets is variable)
  par.ngeof = par.source_index[0];
  par.last_geof = par.first_geof+par.ngeof;

  // Read input data for source estimation
  par.read_shot(0);

  // Initialize Non-linear mod. operator
  NLMod &pnlmod=nlmod;
  NLMod_op op(ppar,pnlmod); 
 
  // Pointer to the operator
  void (Op::*nlmod_op)(int,int,
                 std::vector<float>&,
                 std::vector<float>&)=&Op::op;
  int (Step::*pcgstep)(int&,
                      std::vector<float>&,
                      std::vector<float>&,
                      std::vector<float>&,
                      std::vector<float>&)=&Step::cgstep;

  std::vector<float> model(par.nt);
  std::vector<float> data(par.nkgx*par.nt);

  // Initialize regularization filter
  int (Conv::*ptconv)(int,int,
                     std::vector<float>&,
                     std::vector<float>&)=&Conv::tconv;
  std::vector<float> bc(1,-1);
  bc.push_back(1);
  tconv.tconv_init(bc);
  int nreg=model.size()+bc.size();

  // Initialize model vector 
  for(int jj=0; jj<model.size(); jj++) {
    model[jj]=0;
  }

  // Pass the input data to the data vector 
  for(int jgx=0; jgx<par.nkgx; jgx++) {
    for(int jt=0; jt<par.nt; jt++) {
      int j1=jt+jgx*par.nt;
      int j2=jt+jgx*(par.nt+2);
      data[j1]=par.rcv[j2];
    }
  }

  // A priory model, if any
  std::vector<float> m0;

  // Preconditioner, if any (diag. of GN Hessian)
  std::vector<float> H; // Empty vector as input

  // Vector for the residual
  std::vector<float> resd(data.size());

  // Vector for the residual error
  std::vector<float> err_res(par.niter,0.0);

  // Perform inversion
  std::cout << "Begin linear solver" << std::endl;
  solver.solver_wjreg(model,data,nlmod_op,op,
                      ptconv,tconv,pcgstep,cgstep,m0,H,
                      nreg,par.niter,par.eps,resd,err_res); 
  std::cout << "Finish linear solver" << std::endl;

  // Retrieve the solution from the model vector
  for(int jj=0; jj<model.size(); jj++) {
    par.src[jj]=model[jj];
if(model[jj] != 0.) std::cout << "model : " << model[jj] << endl; 
  }

  std::cout << "*********** MAIN CALCULATION ENDS! ********** " << std::endl;

  // Initialize writing
  out.write_init(par);

  // Write the output
  std::cout << "Write the output binary of modeled source" << std::endl;  
  float **S;
  alloc(S,1,par.nt);
  for(int jt=0; jt<par.nt; jt++) {
    S[0][jt] = par.src[jt];
  }
  par.adj=1;
  out.write_data(par,S,"nlsrc");
  delete S;

  // Write the residual 
  std::cout << "Write the residual binary" << std::endl;  
  float **D;
  alloc(D,par.nkgx,par.nt);
  for(int jgx=0; jgx<par.nkgx; jgx++) {
    for(int jt=0; jt<par.nt; jt++) {
      int jj=jt+jgx*par.nt;
      D[jgx][jt] = resd[jj];
    }
  }
  out.write_data(par,D,"resid");
  delete D;

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "lsqnlmod2d: Finishing successfully!!!" << std::endl;  

  hclose();

  return 0;

}
