/*
  **************************************************************
  *** Main program for 2D least-squares RTM in data space.   ***
  *** Version rewritten in C++                               ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "grad.h"
#include "grad2.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "vshot.h"
#include "rshot.h"
#include <cmath>
#include "snap.h"
#include "alloc.h"
#include "born.h"
#include "rtm.h"
#include "operator.h"
#include "born_op.h"
#include "tconv.h"
#include "solver_wjreg.h"
//#include "solver_cgm.h"
#include "cgstep.h"
#include "dso.h"
//#include "dso_reg.h"
#include <vector>
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <functional>
#include <numeric>

int MAIN() {
  // Define class member objects
  Read par;
  VShot vel;
  WaveProp fwd, scat, bwd;
  Born born; 
  Rtm rtm; 
  Write out;
  Solver solver; 
//  SolverCgm solver; 
  Conv tconv;
  Step cgstep;
  RShot refl;

  std::cout << "******************************************************************" << std::endl;
  std::cout << "*** 2D Least-squares reverse-time migration in data space        *" << std::endl;
  std::cout << "*** Version rewritten in C++                                     *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                   *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017       *" << std::endl;
  std::cout << "******************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Read data header
  par.read_headers();
  
  // Read velocity
  par.read_velocity();

  // Create reflectivity container
  par.read_reflectivity("none");

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Obtain data geometry and apertures from input parameters
  par.data_geometry();
  
  // Read and verbose source
  par.read_src(1);
  
  // Initialize tapering for receiver propagation
  taper_init(par);

  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize some variables for the main loop
  par.ntSnap=0;
  par.first_geof=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Get number of geophone traces in shot gather (in some datasets is variable)
  par.ngeof = par.source_index[0];
  par.last_geof = par.first_geof+par.ngeof;

  // Initialize variables for shot velocities
  vel.vshot_init(par);

  // Initialize parameters for source and scattered wavefields
  fwd.prop_init(par);
  scat.prop_init(par);
  bwd.prop_init(par);

  // Create velocity frames for source wavefield
  vel.vshot(par,par.Vint,fwd.Vshot,par.sgx,1); // Use random boundary cond.

  // Create velocity frames for receiver wavefield
  vel.vshot(par,par.Vint,bwd.Vshot,par.sgx,0); // Don't use random boundary cond.

  // Create velocity frames for scattered source wavefield
  for(int jgx=0; jgx<par.nkgx; jgx++) {
    for(int jz=0; jz<par.nkz; jz++) {
      scat.Vshot[jgx][jz]=fwd.Vshot[jgx][jz];
    }
  }
  
//  vel.vshot(par,par.Vint,scat.Vshot,par.sgx,1); // Use random boundary cond.

  // Initialize Born operator
  Read &ppar=par;
  Born &pborn=born;
  Rtm &prtm=rtm;
  Born_op op(ppar,prtm,pborn,fwd,bwd,scat); 
 
  // Pointer to the operator
  void (Op::*born_op)(int,int,
                 std::vector<float>&,
                 std::vector<float>&)=&Op::op;

  int (Step::*pcgstep)(int&,std::vector<float>&,
                            std::vector<float>&,
                            std::vector<float>&,
                            std::vector<float>&)=&Step::cgstep;

  int n1=par.nkgx*(2*par.nh+1)*par.nkz;
  int n2=par.nkgx*par.nt;
  std::vector<float> model(n1,0.);
  std::vector<float> data(n2,0.);

  // Initialize regularization filter
  int (Conv::*ptconv)(int,int,
                     std::vector<float>&,
                     std::vector<float>&)=&Conv::tconv;

  // Penalty function
  int nreg=0;
  int a=1;
  if(par.eps != 0) { // Use minimum norm or derivative 
    if(par.deriv == 1) a=-1;
    std::vector<float> bc(1,a);
    // deriv=0 min norm; deriv=1 1st derivative
    if(par.deriv ==1 ) bc.push_back(1);
    tconv.tconv_init(bc);
    nreg=model.size()+bc.size();
  }

  // Pass the input data to the data vector 
  data=par.rcv;

  // A priory model, if any
  std::vector<float> m0;

  // Preconditioner, if any (diag. of GN Hessian)
  std::vector<float> H(n1);
  std::vector<float> P(n1);
  if(par.precond==0) {
    P.clear();
    H.clear();
  }
  else {
    par.read_reflectivity("hess");
    refl.rshot_init(par);
    refl.rshot(par,H,par.sgx);
    float H_average=std::accumulate(H.begin(),H.end(),0.0)/H.size(); 
    for(int jj=0; jj<H.size(); jj++) {
      if(H[jj] > 0.) { 
        P[jj]=1.0/sqrt(H[jj]);
      }
      else {
        P[jj]=1.0/sqrt(0.1*H_average);
      }
    }
  }

  // Masking operator, if any
  std::vector<float> J; // Empty vector if not used

  // Data Weighting, if any
  std::vector<float> W(data.size()); // Empty vector if not used
  if(par.weight==1) {
//    par.read_reflectivity("mask");
//    refl.rshot(par,W,par.sgx);      // Don't use random boundary conditions
  }
  else {
    W.clear();
  }

  // Vector for the residual
  std::vector<float> resd(data.size());

  // Vector for the gradient
  std::vector<float> grad(model.size());
  if(par.save_grad == 0) grad.clear();

  // Vector for the residual error
  std::vector<float> err_res(par.niter,0.0);

  // Perform inversion
  std::cout << "Begin linear solver" << std::endl;
  solver.solver_wjreg(model,data,born_op,op,
                        ptconv,tconv,pcgstep,cgstep,
                        m0,P,W,J,nreg,
                        par.niter,par.eps,par.nreboot,
                        resd,grad,err_res); 
//  solver.solver_cgm(model,data,born_op,op,
//                    ptconv,tconv,m0,P,nreg,
//                    par.niter,par.eps,
//                    resd,err_res); 
  std::cout << "Finish linear solver" << std::endl;

  // Retrieve the solution from the model vector
  rtm.num=model;

  // Construct the images
  rtm.build_images(par);

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  std::cout << "Write the inverted reflectivity" << std::endl;  
  out.write_image(par,rtm.ICorr.data(),"rtm_cic");

  // Write the residual 
  std::cout << "Write the residual binary" << std::endl;  
  float **D;
  alloc(D,par.nkgx,par.nt);
  for(int jgx=0; jgx<par.nkgx; jgx++) {
    for(int jt=0; jt<par.nt; jt++) {
      int jj=jt+jgx*par.nt;
      D[jgx][jt] = resd[jj];
    }
  }
  out.write_data(par,D,"resid");
  delete D;

  // Write the gradient
  if(!grad.empty()) {
    // Retrieve the gradient vector
    rtm.num=grad;

    // Construct the gradient images
    rtm.build_images(par);

    std::cout << "Write the final gradient" << std::endl;  
    out.write_image(par,rtm.ICorr.data(),"grad");
  }

  // Write the residual error for plotting
  if(!err_res.empty()) out.write_error_res(par,par.niter,err_res.data(),"error_res");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "lsqrtm2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}

