/*
  **************************************************************
  *** Main program for 2D least-squares RTM in mode space.   ***
  *** Version rewritten in C++                               ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "vshot.h"
#include "rshot.h"
#include <cmath>
#include "snap.h"
#include "alloc.h"
#include "allocv.h"
#include "lint2.h"
#include "operator.h"
#include "hess_op.h"
#include "tconv.h"
#include "solver_wjreg.h"
#include "cgstep.h"
#include <vector>
#include "map.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <functional>
#include <numeric>

int MAIN() {
  // Define class member objects
  Read par;
  Write out;
  RShot rshot;
  Solver solver; 
  Conv tconv; 
  Step cgstep;
  RShot refl;

  std::cout << "******************************************************************" << std::endl;
  std::cout << "*** 2D Least-squares reverse-time migration in model space       *" << std::endl;
  std::cout << "*** Version rewritten in C++                                     *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                   *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017       *" << std::endl;
  std::cout << "******************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Read data header
  par.read_headers();
  
  // Read velocity
  par.read_velocity();

  // Read reflectivity
  par.read_reflectivity("reflfile");

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Obtain data geometry and apertures from input parameters
  par.data_geometry();
  
  // Verbose some useful parameters
  par.verbose_par();
  
  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Initialize model and data vectors
  std::vector<float> model(par.nkgx*(2*par.nh+1)*par.nkz,0.);
  std::vector<float> data (par.nkgx*(2*par.nh+1)*par.nkz,0.);

  // Map reflectivity data to data coordinates
  rshot.rshot_init(par);
  rshot.rshot(par,data,par.sgx);

  // Initialize Hess operator
  Hess_op op(par,1); 

  // Pointer to the operator
  void (Op::*hess_op)(int,int,
                 std::vector<float>&,
                 std::vector<float>&)=&Op::op;
  int (Step::*pcgstep)(int&,
                      std::vector<float>&,
                      std::vector<float>&,
                      std::vector<float>&,
                      std::vector<float>&)=&Step::cgstep;

  // Initialize regularization filter
  int (Conv::*ptconv)(int,int,
                     std::vector<float>&,
                     std::vector<float>&)=&Conv::tconv;

  // Penalty function (deriv=0 min norm; deriv=1 1st derivative)
  int a=1;
  if(par.deriv == 1) a=-1;
  std::vector<float> bc(1,a);
  if(par.deriv ==1 ) bc.push_back(1);
  tconv.tconv_init(bc);
  int nreg=model.size()+bc.size();

  // A priory model, if any. If not, m0 is left empty
  std::vector<float> m0; 

  // Preconditioner, if any
  std::vector<float> P; // Empty vector as input

  // Masking operator, if any
  std::vector<float> J; // Empty vector if not used

  // Data Weighting, if any
  std::vector<float> W(data.size()); // Empty vector if not used
  if(par.weight==1) {
    par.read_reflectivity("mask"); 
    refl.rshot(par,W,par.sgx);      // Don't use random boundary conditions
  }
  else {
    W.clear();
  }

  // Vector for the residual
  std::vector<float> resd(data.size());

  // Vector for the residual error
  std::vector<float> err_res(par.niterH,0.0);

  // Vector for the gradient
  std::vector<float> grad(model.size());
//  if(par.save_grad == 0) grad.clear();
  grad.clear();

  // Perform inversion
  std::cout << "Begin linear solver" << std::endl;
  solver.solver_wjreg(model,data,hess_op,op,
                      ptconv,tconv,pcgstep,cgstep,
                      m0,P,W,J,nreg,
                      par.niterH,par.eps,par.nreboot,
                      resd,grad,err_res); 
  std::cout << "Finish linear solver" << std::endl;

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;
  // Vector for mapping image from data to model coordinates
  std::vector<float> image;
  int n=par.nx*(2*par.nh+1)*par.nz;
  allocV(image,n);

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  std::cout << "Write the reflectivity" << std::endl;  
  map(par,model,image);
  out.write_image(par,image.data(),"lsqrtmh");

  // Write the residual 
  if(!resd.empty()) {
    std::cout << "Write the residual" << std::endl;  
    allocV(image,n);
    map(par,resd,image);
    out.write_image(par,image.data(),"resid");
  }

  // Write the residual error for plotting
  if(!err_res.empty()) out.write_error_res(par,par.niterH,err_res.data(),"error_res");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "lsqrtmH2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}

