/*
  **************************************************************
  *** Main program for 2D least-squares WEMVA.               ***
  *** Version rewritten in C++                               ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "lap.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "vshot.h"
#include "rshot.h"
#include <cmath>
#include "alloc.h"
#include "allocv.h"
#include "wemva.h"
#include "dso.h"
#include "wemva_op.h"
#include "tconv.h"
#include "solver_wjreg.h"
#include "cgstep.h"
#include <vector>
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <algorithm>
#include <functional>
#include <numeric>

int MAIN() {
  // Define class member objects
  Read par;
  VShot vel;
  WaveProp fwd, scat, bwd;
  Wemva wemva,wemva_adj;
  Write out;
  Solver solver; 
  Conv tconv; 
  Step cgstep;
  RShot refl;

  std::cout << "******************************************************************" << std::endl;
  std::cout << "*** 2D Least-squares wave-equation migration velocity analysis   *" << std::endl;
  std::cout << "*** Version rewritten in C++                                     *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                   *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017       *" << std::endl;
  std::cout << "******************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Read data header
  par.read_headers();
  
  // Read velocity
  par.read_velocity();

  // Read migrated image as perturbation in reflectivity
  par.read_pert(1);

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Obtain data geometry and apertures from input parameters
  par.data_geometry();
  
  // Read and verbose source
  par.read_src(1);
  
  // Initialize tapering for receiver propagation
  taper_init(par);

  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize some variables for the main loop
  par.ntSnap=0;
  par.first_geof=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Get number of geophone traces in shot gather (in some datasets is variable)
  par.ngeof = par.source_index[0];
  par.last_geof = par.first_geof+par.ngeof;

  // Initialize variables for shot velocities
  vel.vshot_init(par);

  // Initialize parameters for forward and backward propagations
  fwd.prop_init(par);
  bwd.prop_init(par);
  scat.prop_init(par);

  // Create velocity frames
  vel.vshot(par,par.Vint,fwd.Vshot,par.sgx,par.rbc); // Use random boundary conditions
  vel.vshot(par,par.Vint,bwd.Vshot,par.sgx,par.rbc); // Use random boundary conditions
  vel.vshot(par,par.Vint,scat.Vshot,par.sgx,0);      // Don't use random boundary conditions

  // Initialize WEMVA operator
  Read &ppar=par;
  Wemva &pwemva=wemva;
  Wemva &pwemva_adj=wemva_adj;
  Dso dso(par,1);
  Wemva_op op(ppar,pwemva,pwemva_adj,fwd,bwd,scat,dso,1);

  // Pointer to the operator
  void (Op::*wemva_op)(int,int,
                  std::vector<float>&,
                  std::vector<float>&)=&Op::op;

  int (Step::*pcgstep)(int&,std::vector<float>&,
                            std::vector<float>&,
                            std::vector<float>&,
                            std::vector<float>&)=&Step::cgstep;

  // Initialize model amnd data vectors
  int n=par.nkgx*(2*par.nh+1)*par.nkz;
  std::vector<float> model(n,0.);
  std::vector<float> data(n,0.);

  // Initialize regularization filter
  int (Conv::*ptconv)(int,int,std::vector<float>&,
                              std::vector<float>&)=&Conv::tconv;

  // Penalty function (deriv=0 min norm; deriv=1 1st derivative)
  int a=1;
  if(par.deriv == 1) a=-1;
  std::vector<float> bc(1,a);
  if(par.deriv ==1 ) bc.push_back(1);
  tconv.tconv_init(bc);
  int nreg=model.size()+bc.size();

  // Pass the input data to the data vector: d=-D I0 
//  dso.op(0,0,wemva.pshot,data);
  data=wemva.pshot;
//  if(par.nh != 0) {
    std::transform(data.begin(),data.end(),data.begin(),
         std::bind1st(std::multiplies<float>(),-1.0));
//  }

  // A priory model, if available
  std::vector<float> m0; // Empty vector if not used

  // Preconditioner, if any
  std::vector<float> P; // Empty vector if not used

  // Masking operator, if any
  std::vector<float> J(model.size()); // Empty vector if not used
  if(par.masking == 1) {
   std::cout << "Using mask" << std::endl;
    par.read_reflectivity("mask");
    refl.rshot_init(par);
    refl.rshot(par,J,par.sgx);      // Don't use random boundary conditions
  }
  else {
    J.clear();
  }

  // Data Weighting, if any
  std::vector<float> W; // Empty vector if not used

  // Vector for the residual
  std::vector<float> resd(data.size());

  // Vector for the gradient
  std::vector<float> grad(model.size());
  if(par.save_grad == 0) grad.clear();

  // Vector for the residual error
  std::vector<float> err_res(par.niter,0.0);

  // Perform inversion
  std::cout << "Begin linear solver" << std::endl;
  solver.solver_wjreg(model,data,wemva_op,op,
                     ptconv,tconv,pcgstep,cgstep,
                     m0,P,W,J,nreg,
                     par.niter,par.eps,par.nreboot,
                     resd,grad,err_res); 
  std::cout << "Finish linear solver" << std::endl;

  // Retrieve the solution from the model vector
//  if(par.nh == 0) {
//    std::transform(model.begin(),model.end(),wemva.pert.begin(),
//         std::bind1st(std::multiplies<float>(),-1.0));
//  }
//  else {
    wemva.pert=model;
//  }

  // Construct the images
  wemva.build_pert(par);

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  std::cout << "Write the inverted perturbation in the image" << std::endl;  
  out.write_pert(par,wemva.IPert.data(),"wemva");
  
  // Write the residual 
  if(!resd.empty()) {
    std::cout << "Write the residual" << std::endl;  
    wemva.pert.swap(resd);
    wemva.build_pert(par);
    out.write_image(par,wemva.IPert.data(),"resid");
  }

  // Write the gradient
  if(!grad.empty()) {
    // Retrieve the gradient vector
    wemva.pert=grad;

    // Construct the gradient images
    wemva.build_pert(par);

    std::cout << "Write the final gradient" << std::endl;  
    out.write_image(par,wemva.IPert.data(),"grad");
  }

  // Write the residual error for plotting
  if(!err_res.empty()) out.write_error_res(par,par.niter,err_res.data(),"error_res");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "lsqwemva2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}
