/*
  **************************************************************
  *** Main program for 2D least-squares RTM in mode space.   ***
  *** Version rewritten in C++                               ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "lap.h"
#include "interp1d.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "vshot.h"
#include "rshot.h"
#include <cmath>
#include "snap.h"
#include "alloc.h"
#include "allocv.h"
#include "lint2.h"
#include "wemva.h"
#include "operator.h"
#include "hess_op.h"
#include "dso.h"
#include "wemva_op.h"
#include "lwivu_op.h"
#include "tconv.h"
#include "tconv.h"
#include "solver_wjreg.h"
#include "cgstep.h"
#include <vector>
#include "map.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <functional>
#include <algorithm>
#include <numeric>

int MAIN() {
  // Define class member objects
  Read par;
  VShot vel;
  WaveProp fwd, scat, bwd;
  Write out;
  RShot refl;
  Wemva wemva, wemva_adj;
  Solver solver; 
  Conv tconv;
  Step cgstep;

  std::cout << "******************************************************************" << std::endl;
  std::cout << "*** 2D Least-squares reverse-time migration in model space       *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                   *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017       *" << std::endl;
  std::cout << "******************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Read data header
  par.read_headers();
  
  // Read velocity
  par.read_velocity();

  // Read reflectivity
  par.read_reflectivity("reflfile");

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Obtain data geometry and apertures from input parameters
  par.data_geometry();
  
  // Read and verbose source
  par.read_src(1);
  
  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize some variables for the main calculation
  par.ntSnap=0;
  par.first_geof=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Get number of geophone traces in shot gather (in some datasets is variable)
  par.ngeof = par.source_index[0];
  par.last_geof = par.first_geof+par.ngeof;

  // Initialize variables for shot velocities
  vel.vshot_init(par);

  // Initialize parameters for source and scattered wavefields
  fwd.prop_init(par);
  scat.prop_init(par);
  bwd.prop_init(par);

  // Create velocity frames for source wavefield
  vel.vshot(par,par.Vint,fwd.Vshot,par.sgx,1); // Use random boundary cond.

  // Create velocity frames for receiver wavefield
  vel.vshot(par,par.Vint,bwd.Vshot,par.sgx,0); // Don't use random boundary cond.

  // Create velocity frames for scattered source wavefield
  for(int jgx=0; jgx<par.nkgx; jgx++) {
    for(int jz=0; jz<par.nkz; jz++) {
      scat.Vshot[jgx][jz]=fwd.Vshot[jgx][jz];
    }
  }
  
  // Initialize model and data vectors
  int n=par.nkgx*(2*par.nh+1)*par.nkz;
  std::vector<float> model(2*n,0.);
  std::vector<float> data;
  std::vector<float> temp1(n,0.);

  // Map reflectivity data to data coordinates
  refl.rshot_init(par);
  refl.rshot(par,temp1,par.sgx);

  // Initialize Hess operator
  Hess_op hess_op(par,1); 
  Dso dso(par,1);
  Dso dso_local(par,0);
  Wemva_op wemva_op(par,wemva,wemva_adj,fwd,bwd,scat,dso,1); 
  Lwivu_op op(par,hess_op,wemva_op,dso_local,1);

  // Insert into data
  data.insert(data.end(),temp1.begin(),temp1.end());

  // Compute D data2
//  par.read_reflectivity("reflfile");
//  rshot.rshot_init(par);
//  rshot.rshot(par,temp1,par.sgx);
//  dso_local.op(0,0,wemva.pshot,temp1);
  std::transform(temp1.begin(),temp1.end(),temp1.begin(),
  std::bind1st(std::multiplies<float>(),-1.));

  // Insert into data
  data.insert(data.end(),temp1.begin(),temp1.end());
  temp1.clear();

  // Pointer to the operator
  void (Op::*lwivu_op)(int,int,
                 std::vector<float>&,
                 std::vector<float>&)=&Op::op;
  int (Step::*pcgstep)(int&,
                      std::vector<float>&,
                      std::vector<float>&,
                      std::vector<float>&,
                      std::vector<float>&)=&Step::cgstep;

  // Initialize regularization filter
  int (Conv::*ptconv)(int,int,
                     std::vector<float>&,
                     std::vector<float>&)=&Conv::tconv;

  // Penalty function (deriv=0 min norm; deriv=1 1st derivative)
  int a=1;
  if(par.deriv == 1) a=-1;
  std::vector<float> bc(1,a);
  if(par.deriv ==1 ) bc.push_back(1);
  tconv.tconv_init(bc);
  int nreg=model.size()+bc.size();

  // A priory model, if any. If not, m0 is left empty
  std::vector<float> m0; 

  // Preconditioner, if any
  std::vector<float> P; // Empty vector as input

  // Masking operator, if any
  std::vector<float> J; // Empty vector if not used

  // Data Weighting, if any
  std::vector<float> W(data.size()); // Empty vector if not used
  if(par.weight==1) {
    par.read_reflectivity("mask"); 
    refl.rshot(par,W,par.sgx);      // Don't use random boundary conditions
  }
  else {
    W.clear();
  }

  // Vector for the residual
  std::vector<float> resd(data.size());

  // Vector for the gradient
  std::vector<float> grad(model.size());
//  if(par.save_grad == 0) grad.clear();
  grad.clear();

  // Vector for the residual error
  std::vector<float> err_res(par.niter,0.0);

  // Perform inversion
  std::cout << "Begin linear solver" << std::endl;
  solver.solver_wjreg(model,data,lwivu_op,op,
                      ptconv,tconv,pcgstep,cgstep,
                      m0,P,W,J,nreg,
                      par.niter,par.eps,par.nreboot,
                      resd,grad,err_res); 
  std::cout << "Finish linear solver" << std::endl;

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;
  // Vector for mapping image from data to model coordinates
  std::vector<float> image;
  int n1=par.nx*(2*par.nh+1)*par.nz;
  allocV(image,n1);

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  std::cout << "Write the reflectivity" << std::endl;  
  std::vector<float> split;
  split.insert(split.begin(),model.begin(),model.begin()+n);
  map(par,split,image);
  out.write_image(par,image.data(),"lwivu_refl");

  // Write the residual 
  if(!resd.empty()) {
    std::cout << "Write the residual 1" << std::endl;  
    split.clear();
    image.clear();
    image = std::vector<float> (n1,0);
    split.insert(split.begin(),resd.begin(),resd.end());
    map(par,split,image);
    out.write_image(par,image.data(),"resid1");
  }

  // Vector for mapping image from data to model coordinates
  image.clear();
  image = std::vector<float> (n,0);

  // Vector for splitting the output
  split.clear();

  // Write the output
  std::cout << "Write the perturbation in slowness squared" << std::endl;  
  split.clear();
  split.insert(split.begin(),model.begin()+n,model.end());
  map(par,split,image);
  out.write_image(par,image.data(),"lwivu_pert");

  // Write the residual 
  if(!resd.empty()) {
    std::cout << "Write the residual 2" << std::endl;  
    split.clear();
    split.insert(split.begin(),resd.begin()+n,resd.end());
    map(par,split,image);
    out.write_image(par,image.data(),"resid2");
  }

  // Write the residual error for plotting
  if(!err_res.empty()) out.write_error_res(par,par.niter,err_res.data(),"error_res1");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "lsqrtmH2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}

