#include "read.h"
#include "seplib.h"
#include "vshot.h"
#include "rshot.h"
#include "lap.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "born.h"
#include "rtm.h"
#include "born_op.h"
#include "tconv.h"
#include "cgstep.h"
#include "solver_cgm.h"
#include "solver_report.h"
#include "dot_prodf.h"
#include <vector>
#include <algorithm>
#include <numeric>
#include <iostream>

SolverCgm::~SolverCgm() {}

void SolverCgm::solver_cgm(std::vector<float> &m,
                          std::vector<float> &d,
        void (Op::*pFop)(int,int,std::vector<double>&,
                                 std::vector<double>&),
        Op &Fop,
        int (Conv::*pAop)(int,int,std::vector<float>&,
                                  std::vector<float>&),
        Conv &Aop,
        std::vector<float> &m0,std::vector<float> &P,
        int nAop,int niter,float eps,
        std::vector<float> &resd,std::vector<float> &err_res) {

  std::vector<double> g(m.size(),0.);
  std::vector<double> h(m.size(),0.);
  std::vector<double> R(d.size()+nAop,0.);
  std::vector<double> Q(d.size()+nAop,0.);

  std::vector<double> Rd(d.size(),0.);
  std::vector<double> Rm(nAop,0.);
  std::vector<double> Qd(d.size(),0.);
  std::vector<double> Qm(nAop,0.);

  std::vector<double> epsg(m.size(),0.);
  std::vector<double> epsh(m.size(),0.);
  std::vector<double> epsRm(nAop,0.);
  std::vector<double> epsm0(nAop,0.);

  // Rd = -d
  for(int jj=0; jj<d.size(); jj++) {
    Rd[jj]=d[jj];
  }
  for(int jj=0; jj<nAop; jj++) {
    Rm[jj]=0.0;
  }

//  if(m0.empty()) {
    // Initialize model m=0
    for(int jj=0; jj<m.size(); jj++) {
      m[jj]=0.0;
    }
//  }
//  else {
//    // Initialize model m=m0
//    for(int jj=0; jj<m.size(); jj++) {
//      m[jj]=m0[jj];
//    }
//    (Fop.*pFop)(0,1,m,Rd);
//    std::transform(m0.begin(),m0.end(),epsm0.begin(),
//         std::bind1st(std::multiplies<float>(),eps));
//    if(eps != 0.) (Aop.*pAop)(0,1,epsm0,Rm);
//  }

  /*** Loop over iteration number ***/
  for(int jiter=1; jiter<=niter; jiter++) {
    // g = F'Rd
    (Fop.*pFop)(1,0,g,Rd);

    // eps*Rm 
    std::transform(Rm.begin(),Rm.end(),epsRm.begin(),
         std::bind1st(std::multiplies<float>(),eps));

    // g += A'eps*Rm
//    if(eps != 0.) (Aop.*pAop)(1,1,g,epsRm);

    // g = P'g Preconditioning
    if(!P.empty()) {
      std::vector<double> temp(g.size());
      temp.swap(g);
      for(int jj=0; jj<g.size(); jj++) {
        g[jj]=P[jj]*temp[jj];
      }
    }

    // g = Pg Preconditioning
    if(!P.empty()) {
      std::vector<double> temp(g.size());
      temp.swap(g);
      for(int jj=0; jj<g.size(); jj++) {
        g[jj]=P[jj]*temp[jj];
      }
    }

    // beta=||gnew||^2/||gold||^2
    gdg=dot_prodf(g,g);
    double beta;
    if(jiter==1) {
      beta = 0.;
    }
    else {
      beta=gdg/gdgold;
    }

    // h = -g+beta*h
    for(int jj=0; jj<h.size(); jj++) {
      h[jj]=g[jj]+beta*h[jj];
    } 

    // Qd = Fh
    (Fop.*pFop)(0,0,h,Qd);

    // eps*h 
    std::transform(h.begin(),h.end(),
    epsh.begin(),std::bind1st(std::multiplies<float>(),eps));
    double epshdh=dot_prodf(epsh,epsh);

    // Qm = eps*Ah
//    if(eps != 0.) (Aop.*pAop)(0,0,epsh,Qm);

    // Concatenate vectors
    for(int jj=0; jj<Rd.size(); jj++) {
      R[jj] = Rd[jj];
      Q[jj] = Qd[jj];
    }
    for(int jj=0; jj<Rm.size(); jj++) {
      R[jj+Rd.size()] = Rm[jj];
      Q[jj+Rd.size()] = Qm[jj];
    }
    
    // alpha=||g||^2/||Q||^2
    double QdQ=dot_prodf(Q,Q);
    double alpha=gdg/QdQ;

    // m += alpha*h; R += alpha*Q
    for(int jj=0; jj<m.size(); jj++) {
      m[jj]+=alpha*h[jj];
    }
    for(int jj=0; jj<R.size(); jj++) {
      R[jj]-=alpha*Q[jj];
    }

    // Obtain data residual, gradient, and model residual
    RdR=dot_prodf(Rd,Rd);
    gdgold=gdg;

std::cout << "RdR: " << RdR << std::endl;
std::cout << "alpha: " << alpha << std::endl;
std::cout << "beta: " << beta << std::endl;

    if(RdR < 1e-30) {
      seperr("solver_wjreg: Residual norm is too small! Stop iterating");
      break;
    }
    if(gdg < 1e-30) {
      seperr("solver_wjreg: Gradient norm is too small! Stop iterating");
      break;
    }

    // Save residual error for plotting
    if(!err_res.empty()) err_res[jiter-1]=RdR;

    // Split vectors
    for(int jj=0; jj<Rd.size(); jj++) {
      Rd[jj] = R[jj];
      Qd[jj] = Q[jj];
    }
    for(int jj=0; jj<Rm.size(); jj++) {
      Rm[jj] = R[jj+Rd.size()]; 
      Qm[jj] = Q[jj+Rd.size()]; 
    }
    solver_report(jiter,m,g,Rd,Rm); 

    // Estimate epsilon
//    if(eps == 1.0) {
//      std::vector<float> Ag(g.size());
//      // Ag = A g
//      (Aop.*pAop)(0,0,g,Ag);
//
//      // Estimate epsilon according to Biondi & Barnier (SEP168) 
//      double AgdAg=dot_prodf(Ag,Ag);
//      double eps_est1=sqrt(GdG/AgdAg);
//
//      std::cout << "Estimated epsilon: " << eps_est1 << std::endl;
//      eps=std::min(eps_est1,eps_est2);
//    }
  } // ends number of iterations

  // Preconditioning: Retrieving m = Pinv z
  if(!P.empty()) {
    std::vector<float> temp(m.size());
    temp.swap(m);
    for(int jj=0; jj<m.size(); jj++) {
      m[jj]=temp[jj]/P[jj];
    }
  }

//  if(!resd.empty()) {
//    resd.swap(Rd);
//  }

  std::cout << "*********** MAIN LOOP FINISHES! **********" << std::endl;
  return;
}

