#include "read.h"
#include "seplib.h"
#include "vshot.h"
#include "rshot.h"
#include "lap.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "born.h"
#include "rtm.h"
#include "born_op.h"
#include "tconv.h"
#include "cgstep.h"
#include "solver_wjreg.h"
#include "solver_report.h"
#include "dot_prodf.h"
#include <vector>
#include <algorithm>
#include <numeric>
#include <iostream>

Solver::~Solver() {}

void Solver::solver_wjreg(std::vector<float> &m,
                          std::vector<float> &d,
        void (Op::*pFop)(int,int,std::vector<float>&,
                                 std::vector<float>&),
        Op &Fop,
        int (Conv::*pAop)(int,int,std::vector<float>&,
                                  std::vector<float>&),
        Conv &Aop,
        int (Step::*pcgstep)(int&,std::vector<float>&,
                                  std::vector<float>&,
                                  std::vector<float>&,
                                  std::vector<float>&),
        Step &cgstep,
        std::vector<float> &m0,std::vector<float> &P,
        std::vector<float> &W,std::vector<float> &J,
        int nAop,int niter,float eps,int nreboot,
        std::vector<float> &resd,std::vector<float> &grad,
        std::vector<float> &err_res) {

  std::vector<float> g(m.size(),0.);
  std::vector<float> zero(m.size(),0.);
  std::vector<float> Zero(d.size(),0.);
//  std::vector<float> G(d.size()+nAop,0.);
//  std::vector<float> R(d.size()+nAop,0.);
  std::vector<float> G;
  std::vector<float> R;

  std::vector<float> Rd(d.size(),0.);
  std::vector<float> Rm(nAop,0.);
  std::vector<float> Gd(d.size(),0.);
  std::vector<float> Gm(nAop,0.);

  std::vector<float> epsg(m.size(),0.);
  std::vector<float> epsRm(nAop,0.);
  std::vector<float> epsm0(nAop,0.);

  std::vector<float> temp(g.size());
  std::vector<float> Rtemp(d.size());

  // Rd = -d
  for(int jj=0; jj<d.size(); jj++) {
    Rd[jj]=-d[jj];
  }
  if(!W.empty()) {
    Rtemp.swap(Rd);
    for(int jj=0; jj<Rd.size(); jj++) {
      Rd[jj]=W[jj]*Rtemp[jj];
    }
  }

  for(int jj=0; jj<nAop; jj++) {
    Rm[jj]=0.0;
  }

  if(m0.empty()) {
    // Initialize model m=0
    for(int jj=0; jj<m.size(); jj++) {
      m[jj]=0.0;
    }
  }
  else {
    // Initialize model m=m0
    for(int jj=0; jj<m.size(); jj++) {
      m[jj]=m0[jj];
    }
    (Fop.*pFop)(0,1,m,Rd);
    std::transform(m0.begin(),m0.end(),epsm0.begin(),
         std::bind1st(std::multiplies<float>(),epst));
    if(eps != 0.) (Aop.*pAop)(0,1,epsm0,Rm);
  }

  forget=1;

  /*** Loop over iteration number ***/
  for(int jiter=1; jiter<=niter; jiter++) {
    epst=eps;

    // Restart the steepest-descent method
    if(remainder(jiter,nreboot) == 0) {
      forget=1;
    } 

    // g = F'Rd
    if(!W.empty()) {
      Rtemp.swap(Rd);
      for(int jj=0; jj<Rd.size(); jj++) {
        Rd[jj]=W[jj]*Rtemp[jj];
      }
    }
    (Fop.*pFop)(1,0,g,Rd);

    // Masking g = Jg
    if(!J.empty()) {
      temp.swap(g);
      for(int jj=0; jj<g.size(); jj++) {
        g[jj]=J[jj]*temp[jj];
      }
    }

    // g = P'g Preconditioning
    if(!P.empty()) {
      temp.swap(g);
      for(int jj=0; jj<g.size(); jj++) {
        g[jj]=P[jj]*temp[jj];
      }
    }

    if(eps != 0) {
      // eps*Rm 
      std::transform(Rm.begin(),Rm.end(),epsRm.begin(),
           std::bind1st(std::multiplies<float>(),epst));
  
      // g += A'eps*Rm
      (Aop.*pAop)(1,1,g,epsRm);
    }

    // g = Pg Preconditioning
    if(!P.empty()) {
      for(int jj=0; jj<g.size(); jj++) {
        temp[jj]=P[jj]*g[jj];
      }
    }
    else {
      temp=g;
    }

    // Gd = Fg
    (Fop.*pFop)(0,0,temp,Gd);
    if(!W.empty()) {
      Rtemp.swap(Gd);
      for(int jj=0; jj<Gd.size(); jj++) {
        Gd[jj]=W[jj]*Rtemp[jj];
      }
    }

    if(eps != 0) {
      // eps*g 
      std::transform(g.begin(),g.end(),
      epsg.begin(),std::bind1st(std::multiplies<float>(),epst));
      double epsgdg=dot_prodf(epsg,epsg);

      // Gm = eps*Ag
      (Aop.*pAop)(0,0,epsg,Gm);
    }

    // Concatenate vectors
    R.insert(R.end(),Rd.begin(),Rd.end());    
    G.insert(G.end(),Gd.begin(),Gd.end());    
    if(eps != 0.) {
      R.insert(R.end(),Rm.begin(),Rm.end());    
      G.insert(G.end(),Gm.begin(),Gm.end());
    }
    
    // m += dm; R += dR
    if(1 == (cgstep.*pcgstep)(forget,m,g,R,G))
      seperr("solver_wjreg: Got stuck descending!");

    // Obtain data residual, gradient, and model residual
    RdR=dot_prodf(Rd,Rd);
    gdg=dot_prodf(g,g);
    GdG=dot_prodf(G,G);

    if(RdR < 1e-30) {
      seperr("solver_wjreg: Residual norm is too small! Stop iterating");
      break;
    }
    if(gdg < 1e-30) {
      seperr("solver_wjreg: Gradient norm is too small! Stop iterating");
      break;
    }
    if(GdG < 1e-30) {
      seperr("solver_wjreg: Residual update norm is too small! Stop iterating");
      break;
    }

    // Save residual error for plotting
    if(jiter == 1) RdR0=RdR;
    if(!err_res.empty()) err_res[jiter-1]=RdR/RdR0;

    // Split vectors
    Rd.clear();
    Gd.clear();
    Rm.clear();
    Gm.clear();
    std::vector<float>::iterator it1, it2;
    it1=R.begin()+d.size()-1;
    Rd.insert(Rd.end(),R.begin(),it1);
    it1=G.begin()+d.size()-1;
    Gd.insert(Gd.end(),G.begin(),it1);
    if(epst != 0.) {
      it2=R.begin()+d.size();
      Rm.insert(Rm.end(),it2,R.end());
      it2=G.begin()+d.size();
      Gm.insert(Gm.end(),it2,G.end());
    }
//    for(int jj=0; jj<Rd.size(); jj++) {
//      Rd[jj] = R[jj];
//      Gd[jj] = G[jj];
//    }
//    for(int jj=0; jj<Rm.size(); jj++) {
//      Rm[jj] = R[jj+Rd.size()]; 
//      Gm[jj] = G[jj+Rd.size()]; 
//    }
    R.clear();
    G.clear();

    // Report results
    solver_report(jiter,m,g,Rd,Rm); 

    // Estimate epsilon
    if(eps == 1.0) {
      std::vector<float> Ag(g.size());
      // Ag = A g
      (Aop.*pAop)(0,0,g,Ag);

      // Estimate epsilon according to Biondi & Barnier (SEP168) 
      double AgdAg=dot_prodf(Ag,Ag);
      double eps_est1=sqrt(GdG/AgdAg);

      // Jon's proposal (GIEE eq 3.22)
      double RDdRD=dot_prodf(Rd,Rd);
      double RmdRm=dot_prodf(Rm,Rm);
      double eps_est2=sqrt(RDdRD)/sqrt(RmdRm);

      // Jon's proposal (GIEE eq 3.25)
//      std::vector<float> ARm(Rm.size());
//      (Aop.*pAop)(0,0,Rm,ARm);
//      double gDdgD=std::inner_product(gd.begin(),gd.end(),gd.begin(),0.0);
//      double ARmdARm=std::inner_product(ARm.begin(),ARm.end(),ARm.begin(),0.0);
//      double eps_est1=sqrt(gDdgD/ARmdARm);

      std::cout << "Estimated epsilon: " << eps_est1 << std::endl;
      eps=std::min(eps_est1,eps_est2);
    }
  } // ends number of iterations

  // Preconditioning: Retrieving m = Pinv z
  if(!P.empty()) {
    temp.swap(m);
    for(int jj=0; jj<m.size(); jj++) {
      m[jj]=temp[jj]*P[jj];
    }
  }

  // Get the residual data
  if(!resd.empty()) {
    resd.swap(Rd);
  }

  // Get the gradient image
  if(!grad.empty()) {
    grad.swap(g);
  }

  std::cout << "*********** MAIN LOOP FINISHES! **********" << std::endl;
  return;
}

