/*
  ********************************************************************
  *** B-spline module adapted to C++ from  Y. Tang thesis (sep144) ***
  *** Author: Alejandro Cabrales                                   ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2018       ***
  ********************************************************************
 */ 
#include <vector>
#include "base_geom.h"
#include "base_spline.h"
#include <cmath>
#include <algorithm>

void Spline::base_spln1d_init(Axis axd_in,
                              Axis axk_in) {
  axd=axd_in;
  axk=axk_in;
}

void Spline::base_spln2d_init(Axis axd_in,Axis ayd_in,
                              Axis axk_in,Axis ayk_in) {
  axd=axd_in;
  axk=axk_in;
  ayd=ayd_in;
  ayk=ayk_in;
}

void Spline::base_spln3d_init(Axis axd_in,Axis ayd_in,Axis azd_in,
                              Axis axk_in,Axis ayk_in,Axis azk_in) {
  axd=axd_in;
  axk=axk_in;
  ayd=ayd_in;
  ayk=ayk_in;
  azd=azd_in;
  azk=azk_in;
}


void Spline::base_spln1d(int adj,int add,std::vector<float> &model,
                                         std::vector<float> &data) {
  if(adj==1) {
    if(add==0) {
      for(int jj=0; jj< model.size(); jj++) {
        model[jj]=0.;
      }
    }
    for(int ixd=0; ixd<axd.n; ixd++) {
      float xd=axd.o+ixd*axd.d;
      int jxd=(xd-axk.o)/axk.d;
      for(int ixk=std::max(0,jxd); ixk<std::min(axk.n,jxd+1); ixk++) {
        float xk=axk.o+ixk*axk.d;
        int jm=ixk;
        int jd=ixd;
        model[jm]+=data[jd]*spline(xd-xk,axk.d);
      }
    }
  }
  else {
    if(add==0) {
      for(int jj=0; jj< data.size(); jj++) {
        data[jj]=0.;
      }
    }
    for(int ixd=0; ixd<axd.n; ixd++) {
      float xd=axd.o+ixd*axd.d;
      int jxd=(xd-axk.o)/axk.d;
      for(int ixk=std::max(0,jxd); ixk<std::min(axk.n,jxd+1); ixk++) {
        float xk=axk.o+ixk*axk.d;
        int jm=ixk;
        int jd=ixd;
        data[jd]+=model[jm]*spline(xd-xk,axk.d);
      }
    }
  }
}

void Spline::base_spln2d(int adj,int add,std::vector<float> &model,
                                         std::vector<float> &data) {
  if(adj==1) {
    if(add==0) {
      for(int jj=0; jj< model.size(); jj++) {
        model[jj]=0.;
      }
    }
    for(int iyd=0; iyd<ayd.n; iyd++) {
      float yd=ayd.o+iyd*ayd.d;
      int jyd=(int) (yd-ayk.o)/ayk.d;
      for(int ixd=0; ixd<axd.n; ixd++) {
        float xd=axd.o+ixd*axd.d;
        int jxd=(int) (xd-axk.o)/axk.d;
        for(int iyk=std::max(0,jyd); iyk<std::min(ayk.n,jyd+2); iyk++) {
//        for(int iyk=0; iyk<ayk.n; iyk++) {
          float yk=ayk.o+iyk*ayk.d;
          for(int ixk=std::max(0,jxd); ixk<std::min(axk.n,jxd+2); ixk++) {
//          for(int ixk=0; ixk<axk.n; ixk++) {
            float xk=axk.o+ixk*axk.d;
            int jm=ixk+iyk*axk.n;
            int jd=ixd+iyd*axd.n;
            model[jm]+=data[jd]*spline(xd-xk,axk.d)*spline(yd-yk,ayk.d);
          }
        }
      }
    }
  }
  else {
    if(add==0) {
      for(int jj=0; jj< data.size(); jj++) {
        data[jj]=0.;
      }
    }
    for(int iyd=0; iyd<ayd.n; iyd++) {
      float yd=ayd.o+iyd*ayd.d;
      int jyd=(int) (yd-ayk.o)/ayk.d;
      for(int ixd=0; ixd<axd.n; ixd++) {
        float xd=axd.o+ixd*axd.d;
        int jxd=(int) (xd-axk.o)/axk.d;
        for(int iyk=std::max(0,jyd); iyk<std::min(ayk.n,jyd+2); iyk++) {
//        for(int iyk=0; iyk<ayk.n; iyk++) {
          float yk=ayk.o+iyk*ayk.d;
          for(int ixk=std::max(0,jxd); ixk<std::min(axk.n,jxd+2); ixk++) {
//          for(int ixk=0; ixk<axk.n; ixk++) {
            float xk=axk.o+ixk*axk.d;
            int jm=ixk+iyk*axk.n;
            int jd=ixd+iyd*axd.n;
            data[jd]+=model[jm]*spline(xd-xk,axk.d)*spline(yd-yk,ayk.d);
          }
        }
      }
    }
  }
}

void Spline::base_spln3d(int adj,int add,std::vector<float> &model,
                                         std::vector<float> &data) {
  if(adj==1) {
    if(add==0) {
      for(int jj=0; jj< model.size(); jj++) {
        model[jj]=0.;
      }
    }
    for(int izd=0; izd<azd.n; izd++) {
      float zd=azd.o+izd*azd.d;
      int jzd=(zd-azk.o)/azk.d;
      for(int iyd=0; iyd<ayd.n; iyd++) {
        float yd=ayd.o+iyd*ayd.d;
        int jyd=(yd-ayk.o)/ayk.d;
        for(int ixd=0; ixd<axd.n; ixd++) {
          float xd=axd.o+ixd*axd.d;
          int jxd=(xd-axk.o)/axk.d;
          for(int izk=std::max(0,jzd-1); izk<std::min(azk.n,jzd+2); izk++) {
            float zk=azk.o+izk*azk.d;
            for(int iyk=std::max(0,jyd-1); iyk<std::min(ayk.n,jyd+2); iyk++) {
              float yk=ayk.o+iyk*ayk.d;
              for(int ixk=std::max(0,jxd-1); ixk<std::min(axk.n,jxd+2); ixk++) {
                float xk=axk.o+ixk*axk.d;
                int jm=ixk+iyk*axk.n+izk*ayk.n*azk.n;
                int jd=ixd+iyd*axd.n+izd*ayd.n*azd.n;
                model[jm]+=data[jd]*spline(xd-xk,axk.d)*spline(yd-yk,ayk.d)*spline(zd-zk,azk.d);
              }
            }
          }
        }
      }
    }
  }
  else {
    if(add==0) {
      for(int jj=0; jj< data.size(); jj++) {
        data[jj]=0.;
      }
    }
    for(int izd=0; izd<azd.n; izd++) {
      float zd=azd.o+izd*azd.d;
      int jzd=(zd-azk.o)/azk.d;
      for(int iyd=0; iyd<ayd.n; iyd++) {
        float yd=ayd.o+iyd*ayd.d;
        int jyd=(yd-ayk.o)/ayk.d;
        for(int ixd=0; ixd<axd.n; ixd++) {
          float xd=axd.o+ixd*axd.d;
          int jxd=(xd-axk.o)/axk.d;
          for(int izk=std::max(0,jzd-1); izk<std::min(azk.n,jzd+2); izk++) {
            float zk=azk.o+izk*azk.d;
            for(int iyk=std::max(0,jyd-1); iyk<std::min(ayk.n,jyd+2); iyk++) {
              float yk=ayk.o+iyk*ayk.d;
              for(int ixk=std::max(0,jxd-1); ixk<std::min(axk.n,jxd+2); ixk++) {
                float xk=axk.o+ixk*axk.d;
                int jm=ixk+iyk*axk.n+izk*ayk.n*azk.n;
                int jd=ixd+iyd*axd.n+izd*ayd.n*azd.n;
                data[jd]+=model[jm]*spline(xd-xk,axk.d)*spline(yd-yk,ayk.d)*spline(zd-zk,azk.d);
              }
            }
          }
        }
      }
    }
  }
}

float Spline::spline(float x,float dx) {
  float spln;
  if(x < -2.*dx) {
    spln = 0.;
  }
  else if(x >= -2.*dx && x < -dx) {
    spln = pow(2.*dx+x,3)/6./pow(dx,3);
  }
  else if(x >= -dx && x < 0) {
    spln = pow(-x,3)/2./pow(dx,3)-pow(-x,2)/pow(dx,2)+2./3.;
  }
  else if(x >= 0 && x < dx) {
    spln = pow(+x,3)/2./pow(dx,3)-pow(+x,2)/pow(dx,2)+2./3.;
  }
  else if(x >= dx && x <= 2.*dx) {
    spln = pow(2.*dx-x,3)/6./pow(dx,3);
  }
  else {
    spln = 0.;
  }
//  spln = spln/dx;
  return spln;
}

