#include "read.h"
#include "interp.h"
#include "grad.h"
#include "grad2.h"
#include "wave_prop.h"
#include "lint1.h"
#include "scale.h"
#include "scatter.h"
#include "taper.h"
#include "snap.h"
#include "alloc.h"
#include <vector>
#include "allocv.h"
#include <cmath>
#include "born.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <iostream>
#include <chrono>

void Born :: born_init(Read par,WaveProp &fwd_in,WaveProp &scat_in) {
  // Shorcuts for frequently object's members invokation
  pad=par.pad;
  zpad=par.zpad;
  nkgx=par.nkgx;
  nkz=par.nkz;
  nxE=par.nxE;
  nzE=par.nzE;
  nt=par.nt;
  napL=par.napL;
  napR=par.napR;
  dt=par.dt;
  dx=par.dx;
  dz=par.dz;
  use_taper=par.use_taper;
  nh=par.nh;
  dh=par.dh;
  hmax=par.hmax;
  datumS=par.datumS;
  datumR=par.datumR;
  shot_pos=par.shot_pos;
  eic=par.eic;
  fwd=&fwd_in; 
  scat=&scat_in; 
  dipole=par.dipole;

  // Allocate arrays for scattered wavefield computations
  allocV(dsynt,nkgx*nt);
  allocV(rshot,nkgx*nkz*(2*nh+1));
  if(eic==1) {
    allocV(paux3,nkgx*nkz);
  }

  // Create reflectivity frames 
  par.refl_frames(rshot);

  // Vertical position of the source (dipole source)
  jsrc1=datumS  +pad+zpad+(pad+napL+shot_pos)*nkz;
  if(dipole) jsrc2=datumS+1+pad+zpad+(pad+napL+shot_pos)*nkz;

  // Initialize first derivative for EIC
//  if(eic==1) grad.grad_init(par);
}

void Born :: born(Read &par) {
  // Initialize/update wavefield arrays
  (*fwd).init_wavefields();
  (*scat).init_wavefields();

  std::vector<float> paux1(nkgx*nkz,0.);
  std::vector<float> paux2(nkgx*nkz,0.);

  // Initialize scalar
  scaleS.scale_init(par,(*fwd).Vel2,0);
  scaleScat.scale_init(par,(*fwd).Vel2,1);

  // Initialize scatter
  scatter.scatter_init(par,(*fwd).pm1);

  std::cout << "Perform Born modeling" << std::endl;
  for(int jt=0; jt<nt; jt++) {
    // Forward propagation of source wavefield
    if(jt>1) (*fwd).prop(0);

    if(jt>0) {
      // Prepare dipole source wavelet
      paux1[jsrc1]= par.src[jt-1];
      if(dipole) paux1[jsrc2]=-par.src[jt-1];

      // Scale, then inject the source
//      if(remainder(jt,par.scatStep)==0)
        scaleS.scale(0,1,paux1,(*fwd).p);

      // Damping function applied to the source wavefield
      if(use_taper == 1) 
        taper((*fwd).pm2,(*fwd).pm1,(*fwd).p);
    }

    // Save snapshots: source wavefield
    snapshot(par,jt,(*fwd).p,1,"snap1");

    if(jt>2) {
      // Propagate the scattered source wavefield
      (*scat).prop(0);
    }

    if(jt>1 && remainder(jt,par.scatStep)==0) {
      // Scatter the source wavefield upon reflectivity (ext. domain)
      scatter.scatter(0,0,rshot,paux2); 

      // Scale the scattered source wavefield before propagation
      scaleScat.scale(0,1,paux2,(*scat).p);
    }
//    else {
//      paux2=(*scat).p;
//      scale.scale(0,0,paux2,(*scat).p);
//    }

      // Damping function applied to the scattered wavefield
//      if(use_taper == 1) 
//        taper((*scat).pm2,(*scat).pm1,(*scat).p);

    // Save snapshots: scattered wavefield
//    snapshot(par,jt,scat.p,0,"snap2");

    // Extract the wavefield at the surface
    #pragma omp parallel for
    for(int jgx=0; jgx<nkgx; jgx++) {
      int j1=datumR  +pad+zpad+jgx*nkz;
      int j2=datumR+1+pad+zpad+jgx*nkz;
      int j3=jt+jgx*nt;
      dsynt[j3]+=(*scat).p[j1];
      if(dipole) dsynt[j3]-=(*scat).p[j2];
    }
  }
  
}

