/*
  **************************************************************
  *** Main program for 2D Born modeling.                     ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include <cmath>
#include "alloc.h"
#include "allocv.h"
#include "born.h"
#include "grad.h"
#include "rtm.h"
#include <vector>
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {
  // Define class member objects
  Read par;
  WaveProp fwd, scat;
  Born born; 
  Write out;
 
  std::cout << "**************************************************************" << std::endl;
  std::cout << "*** 2D Born modeling using finite differences                *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                               *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017   *" << std::endl;
  std::cout << "**************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Read data header
//  par.read_headers();
  
  // Build geomery from input parameters
  par.build_geometry();

  // Read velocity
  par.read_velocity();
 
  // Read reflectivity
  par.read_reflectivity();

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Obtain data geometry and apertures from input parameters
//  par.data_geometry();
  
  // Read and verbose source
  par.read_src(1);

  // Initialize tapering for receiver propagation
  taper_init(par);

  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize some variables for the main loop
  par.ntSnap=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Initialize parameters for source and scattered wavefields
  fwd.prop_init(par);
  scat.prop_init(par);

  // Create velocity frames
  par.vel_frames(fwd.Vshot,par.rbc);
  par.vel_frames(scat.Vshot,par.rbc);

  // Initialize parameters for Born modeling
  born.born_init(par,fwd,scat);

  // Run Born modeling
  born.born(par);

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;

  // Initialize writing
  out.write_init(par);

  // Write the output
  out.write_data(par,born.dsynt,"born");

  // Write the headers
//  out.write_data_headers(par,"born");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "born2d: Finishing successfully!!!" << std::endl;  

  hclose();

  return 0;
} 
