/*
  ****************************************************************
  *** Main program for B-spline smoothing.                     ***
  *** B-spline functions modified from Y. Tang thesis (sep144) ***
  *** Author: Alejandro Cabrales                               ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2018   ***
  ****************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "alloc.h"
#include "allocv.h"
#include <vector>
#include <cmath>
#include "base_geom.h"
#include "base_spline.h"
#include <sep.main>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {

  // Define class member objects
  Read par;
  Axis azd,azk,axd,axk;
  Spline spln;
  Write out;

  std::cout << "***************************************************************" << std::endl;
  std::cout << "*** B-spline smoothing using B-spline interpolation operators *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2018    *" << std::endl;
  std::cout << "***************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read padding parameter (employed in LWIVU)
  int padding;
  if(0 == getch("padding","d",&padding))
    padding=0;
  std::cout << "Choose padding? " << padding << std::endl;

  // Read data history
  par.read_history();
  
  // Verbose some useful parameters
  par.verbose_par();

  // Initialize Axis objects
  float factX=par.factX;
  float factZ=par.factZ;
  int o1=-(par.zpad+par.pad)*par.dz;
  azd.o=o1;      azk.o=o1;
  azd.d=par.dz;  azk.d=par.dz*factZ;
  azd.n=par.nz+par.zpad+2*par.pad; azk.n=1+(int) (azd.n/factZ);
  int o2=-(par.pad+par.napL)*par.dx;
  axd.o=o2;      axk.o=o2;
  axd.d=par.dx;  axk.d=par.dx*factX;
  axd.n=par.nx+2*par.pad+par.napL+par.napR; axk.n=1+(int) (axd.n/factX);

  // Report the spacings
  std::cout << "axk.n: " << axk.n << std::endl;
  std::cout << "azk.n: " << azk.n << std::endl;
  std::cout << "axd.n: " << axd.n << std::endl;
  std::cout << "azd.n: " << azd.n << std::endl;

  // Initialize spline2d
  spln.base_spln2d_init(azd,axd,azk,axk);

  int nd=azd.n*axd.n;
  int nm=azk.n*axk.n;
  std::vector<float> data(nd,0.);
  std::vector<float> model_pad(nd,0.);
  std::vector<float> model(nm,0.);

  if(par.adj == 1) {
    // Read data file
    if(nd*4 != sreed("input",data.data(),nd*4))
      seperr("bspline2d: Cannot read data file");
  }
  else {
    // Read model file
    if(padding == 1) {
      if(nd*4 != sreed("input",model_pad.data(),nd*4))
        seperr("bspline2d: Cannot read model file");
    }
    else {
      if(nm*4 != sreed("input",model.data(),nm*4))
        seperr("bspline2d: Cannot read model file");
    }
  }

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;

  if(par.adj == 1) {
    // Adjoint B-spline
    spln.base_spln2d(1,0,model,data);
    float fact=1./factX/factZ;
//    std::transform(model.begin(), model.end(), model.begin(),
//               std::bind1st(std::multiplies<float>(), fact));
    if(padding == 1) {
      for(int jj=0; jj<model.size(); jj++) {
        model_pad[jj]=model[jj];
      }
    } 
  }
  else {
    // Forward B-spline
    if(padding == 1) {
      for(int jj=0; jj<model.size(); jj++) {
        model[jj]=model_pad[jj];
      }
    } 
    spln.base_spln2d(0,0,model,data);
  }

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  std::cout << "Write the output for bspline2d" << std::endl;  
  if(par.adj == 1 && padding == 0) {
//    out.write_image(par,model.data(),"spline");
    if(0 != auxputch("n1","d",&azk.n,"spline"))
      seperr("bspline2d: Cannot write n1 upon history file"); 
    if(0 != auxputch("n2","d",&axk.n,"spline"))
      seperr("bspline2d: Cannot write n2 upon history file"); 
    int one=1;
    if(0 != auxputch("n3","d",&one,"spline"))
      seperr("bspline2d: Cannot write n3 upon history file"); 
    if(0 != auxputch("d1","f",&azk.d,"spline"))
      seperr("bspline2d: Cannot write d1 upon history file"); 
    if(0 != auxputch("d2","f",&axk.d,"spline"))
      seperr("bspline2d: Cannot write d2 upon history file"); 
    float onef=1.;
    if(0 != auxputch("d3","f",&onef,"spline"))
      seperr("bspline2d: Cannot write d3 upon history file"); 
    if(0 != auxputch("o1","f",&azk.o,"spline"))
      seperr("bspline2d: Cannot write o1 upon history file"); 
    if(0 != auxputch("o2","f",&axk.o,"spline"))
      seperr("bspline2d: Cannot write o2 upon history file"); 
    if(0 != auxputch("o3","f",&onef,"spline"))
      seperr("bspline2d: Cannot write o3 upon history file"); 

     // Save redundant values with different names,
     // for its usage in LWIVU
    int dumm1 = auxputch("n1_bspline_adj","d",&azk.n,"spline");
    int dumm2 = auxputch("n2_bspline_adj","d",&axk.n,"spline");
    int n1_diff=abs(azk.n-azd.n);
    int n2_diff=abs(axk.n-axd.n);
    dumm1 = auxputch("n1_diff_adj","d",&n1_diff,"spline");
    dumm2 = auxputch("n2_diff_adj","d",&n2_diff,"spline");
  }
  else {
//    out.write_image(par,data.data(),"spline");
    if(0 != auxputch("n1","d",&azd.n,"spline"))
      seperr("bspline2d: Cannot write n1 upon history file"); 
    if(0 != auxputch("n2","d",&axd.n,"spline"))
      seperr("bspline2d: Cannot write n2 upon history file"); 
    int one=1;
    if(0 != auxputch("n3","d",&one,"spline"))
      seperr("bspline2d: Cannot write n3 upon history file"); 
    if(0 != auxputch("d1","f",&azd.d,"spline"))
      seperr("bspline2d: Cannot write d1 upon history file"); 
    if(0 != auxputch("d2","f",&axd.d,"spline"))
      seperr("bspline2d: Cannot write d2 upon history file"); 
    float onef=1.;
    if(0 != auxputch("d3","f",&onef,"spline"))
      seperr("bspline2d: Cannot write d3 upon history file"); 
    if(0 != auxputch("o1","f",&azd.o,"spline"))
      seperr("bspline2d: Cannot write o1 upon history file"); 
    if(0 != auxputch("o2","f",&axd.o,"spline"))
      seperr("bspline2d: Cannot write o2 upon history file"); 
    if(0 != auxputch("o3","f",&onef,"spline"))
      seperr("bspline2d: Cannot write o3 upon history file"); 

     // Save redundant values with different names,
     // for its usage in LWIVU
    int dumm1 = auxputch("n1_bspline_fwd","d",&azk.n,"spline");
    int dumm2 = auxputch("n2_bspline_fwd","d",&axk.n,"spline");
    int n1_diff=abs(azk.n-azd.n);
    int n2_diff=abs(axk.n-axd.n);
    dumm1 = auxputch("n1_diff_fwd","d",&n1_diff,"spline");
    dumm2 = auxputch("n2_diff_fwd","d",&n2_diff,"spline");
  }
  if(par.adj == 1) {
    if(padding == 1) {
      if(4*nd != srite("spline",model_pad.data(),4*nd))
        seperr("bspline2d: Have trouble writing trace");
    }
    else {
      if(4*nm != srite("spline",model.data(),4*nm))
        seperr("bspline2d: Have trouble writing trace");
    }
  }
  else {
    if(4*nd != srite("spline",data.data(),4*nd))
      seperr("bspline2d: Have trouble writing trace");
  }

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " 
            << std::chrono::duration<double>(t2-t1).count() 
            << "secs." << std::endl; 
  std::cout << "bspline2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}
