/*
  ***************************************************************
  *** Main program for 2D Differential Semblance Optimization ***
  *** Author: Alejandro Cabrales                              ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2018  ***
  ***************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <cmath>
#include "dso.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <algorithm>
#include <numeric>

int MAIN() {
  // Define class member objects
  Read par;
  Write out;

  std::cout << "*************************************************************" << std::endl;
  std::cout << "*** 2D Differential semblance optimization                  *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                              *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2018  *" << std::endl;
  std::cout << "*************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Enforce DSO extension
  par.dso=1;

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Build geomery from input parameters
  par.build_geometry();

  // Read velocity
//  par.read_velocity();
  
  // Read reflectivity
  par.read_reflectivity();

  // Verbose some useful parameters
  par.verbose_par();
 
  std::cout << "*********** MAIN COMPUTATION BEGINS! ********** " << std::endl;

  // Initialize DSO operator
  Read &ppar=par;
  Dso op(par,1);

  // Initialize model and data vectors
  int nmod=par.nxE*par.nzE*(2*par.nh+1);
  int ndat=par.nxE*par.nzE*(2*par.nh+1);
  std::vector<float> model(nmod,0.);
  std::vector<float> data(ndat,0.);

  if(par.adj==1) {
  // Adjoint DSO 
    data.swap(par.refl);
    op.op(1,0,model,data);
  }
  else {
  // Forward DSO
    model.swap(par.refl);
    op.op(0,0,model,data);
  } 

  std::cout << "*********** MAIN COMPUTATION FINISHES! ********** " << std::endl;

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  std::cout << "Write the output for DSO" << std::endl;  
  if(par.adj == 1) {
    if(par.neg_dso == 1) std::transform(model.begin(),model.end(),model.begin(),std::negate<float>());
    out.write_image(par,model.data(),"dso_image");
  }
  else {
    if(par.neg_dso == 1) std::transform(data.begin(),data.end(),data.begin(),std::negate<float>());
    out.write_image(par,data.data(),"dso_image");
  }
  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "dso2d: Finishing successfully!!!" << std::endl;  

  hclose();

 return 0; 
}

