#include "read.h"
#include "interp1d.h"
#include <vector>
#include "grad.h"
#include "allocv.h"
#include <iostream>
#include <tbb/tbb.h>
using namespace tbb;

void Grad::grad_init(Read par) {
  // Some shortcuts
  nkgx=par.nkgx;
  nkz =par.nkz;
  nkgxEx=nkgx+2*buff;
  nkzEx =nkz +2*buff;
  dx  =par.dx;
  dz  =par.dz;

  // Coefficients of 1st deriv., 12th order approx.
  c1 =  160083.0/131072.0; 
  c2 = -12705.0/131072.0 ;
  c3 =  22869.0/1310720.0;
  c4 = -5445.0/1835008.0 ;
  c5 =  847.0/2359296.0  ;
  c6 = -63.0/2883584.0   ;

  // Create coefficients with the spacings 
  float idx=1.0/dx;
  float idz=1.0/dz;
  b1x=c1*idx;
  b2x=c2*idx;
  b3x=c3*idx;
  b4x=c4*idx;
  b5x=c5*idx;
  b6x=c6*idx;
                
  b1z=c1*idz;
  b2z=c2*idz;
  b3z=c3*idz;
  b4z=c4*idz;
  b5z=c5*idz;
  b6z=c6*idz;

  // Allocate wavefield arrays
  allocV(dpx_cur,nkgxEx*nkzEx);
  allocV(dpz_cur,nkgxEx*nkzEx);

  // Allocate auxiliary arrays
  allocV(paux1,nkz);
  allocV(paux2,nkgx);
  allocV(paux3,nkgx);
  allocV(paux4,nkz);
}

void Grad::gradx(std::vector<float>&p,
                std::vector<float>&pX,int adj) {
  // Initialize dpx_cur and dpz to zero
  #pragma omp parallel for
  for(int jj=0; jj<dpx_cur.size(); jj++) {
    dpx_cur[jj]=0;
  }

  #pragma omp parallel for
  for(int jgx=buff; jgx<nkgx+buff; jgx++) {
    #pragma omp simd
    for(int jz=buff; jz<nkz+buff; jz++) {
      int jj=jz+jgx*nkzEx;
      int ii=jz-buff+(jgx-buff)*nkz;
      dpx_cur[jj]=p[ii];
      pX[ii]=0.0;
    }
  }
  if(adj==1) {
    // Interpolate from the staggered grid xs
//    #pragma omp parallel for private(paux2,paux3)
    for(int jz=0; jz<nkz; jz++) {
      #pragma omp simd
      for(int jgx=0; jgx<nkgx; jgx++) {
        int jj=jz+buff+(jgx+buff)*nkzEx;
        paux2[jgx]=dpx_cur[jj];
      }
      interp1d(1,0,paux3,paux2,nkgx);
      #pragma omp simd
      for(int jgx=0; jgx<nkgx; jgx++) {
        int jj=jz+buff+(jgx+buff)*nkzEx;
        dpx_cur[jj]=paux3[jgx];
      }
    }

    // Compute the first derivative in x
    fddx(pX,1);
  }
  else {
    // Compute the first derivative in x
    fddx(pX,0);

    // Interpolate from the staggered grid xs
//    #pragma omp parallel for private(paux2,paux3)
    for(int jz=0; jz<nkz; jz++) {
      #pragma omp simd
      for(int jgx=0; jgx<nkgx; jgx++) {
        int jj=jz+jgx*nkz;
        paux2[jgx]=pX[jj];
      }
      interp1d(0,0,paux2,paux3,nkgx);
      #pragma omp simd
      for(int jgx=0; jgx<nkgx; jgx++) {
        int jj=jz+jgx*nkz;
        pX[jj]=paux3[jgx];
      }
    }
  }
}

void Grad::gradz(std::vector<float>&p,
                std::vector<float>&pZ,int adj) {
  // Initialize dpx_cur and dpz to zero
  #pragma omp parallel for
  for(int jj=0; jj<dpx_cur.size(); jj++) {
    dpz_cur[jj]=0;
  }

  #pragma omp parallel for
  for(int jgx=buff; jgx<nkgx+buff; jgx++) {
    #pragma omp simd
    for(int jz=buff; jz<nkz+buff; jz++) {
      int jj=jz+jgx*nkzEx;
      int ii=jz-buff+(jgx-buff)*nkz;
      dpz_cur[jj]=p[ii];
      pZ[ii]=0.0;
    }
  }
  if(adj==1) {
    // Interpolate from the staggered grid xs
//    #pragma omp parallel for private(paux1)
    for(int jgx=0; jgx<nkgx; jgx++) {
      #pragma omp simd
      for(int jz=0; jz<nkz; jz++) {
        int jj=jz+buff+(jgx+buff)*nkzEx;
        paux1[jz]=dpz_cur[jj];
      }
      interp1d(1,0,paux4,paux1,nkz);
     #pragma omp simd
      for(int jz=0; jz<nkz; jz++) {
        int jj=jz+buff+(jgx+buff)*nkzEx;
        dpz_cur[jj]=paux4[jz];
      }
    }

    // Compute the first derivative in z
    fddz(pZ,1);
  }
  else {
    // Compute the first derivative in z
    fddz(pZ,0);

    // Interpolate from the staggered grid xs
//    #pragma omp parallel for private(paux1)
    for(int jgx=0; jgx<nkgx; jgx++) {
     #pragma omp simd
      for(int jz=0; jz<nkz; jz++) {
        int jj=jz+jgx*nkz;
        paux1[jz]=pZ[jj];
      }
      interp1d(0,0,paux1,paux4,nkz);
     #pragma omp simd
      for(int jz=0; jz<nkz; jz++) {
        int jj=jz+jgx*nkz;
        pZ[jj]=paux4[jz];
      }
    }
  }
}

void Grad::fddx(std::vector<float>&dpx,int adj) {
  if(adj==1) {
    #pragma omp parallel for
    for(int jgx=buff; jgx<nkgx+buff; jgx++) {
      #pragma omp simd
      for(int jz=buff; jz<nkz+buff; jz++) {
        int jj=jz+jgx*nkzEx;
        int ii=jz-buff+(jgx-buff)*nkz;
        dpx[ii] =
          b1x*(-dpx_cur[jj        ]+dpx_cur[jj-1*nkzEx])+
          b2x*(-dpx_cur[jj+1*nkzEx]+dpx_cur[jj-2*nkzEx])+
          b3x*(-dpx_cur[jj+2*nkzEx]+dpx_cur[jj-3*nkzEx])+
          b4x*(-dpx_cur[jj+3*nkzEx]+dpx_cur[jj-4*nkzEx])+
          b5x*(-dpx_cur[jj+4*nkzEx]+dpx_cur[jj-5*nkzEx])+
          b6x*(-dpx_cur[jj+5*nkzEx]+dpx_cur[jj-6*nkzEx]);
      }
    }
  }
  else {
    #pragma omp parallel for
    for(int jgx=buff; jgx<nkgx+buff; jgx++) {
      #pragma omp simd
      for(int jz=buff; jz<nkz+buff; jz++) {
        int jj=jz+jgx*nkzEx;
        int ii=jz-buff+(jgx-buff)*nkz;
        dpx[ii] =
          b1x*(dpx_cur[jj+1*nkzEx]-dpx_cur[jj        ])+
          b2x*(dpx_cur[jj+2*nkzEx]-dpx_cur[jj-1*nkzEx])+
          b3x*(dpx_cur[jj+3*nkzEx]-dpx_cur[jj-2*nkzEx])+
          b4x*(dpx_cur[jj+4*nkzEx]-dpx_cur[jj-3*nkzEx])+
          b5x*(dpx_cur[jj+5*nkzEx]-dpx_cur[jj-4*nkzEx])+
          b6x*(dpx_cur[jj+6*nkzEx]-dpx_cur[jj-5*nkzEx]);
      }
    }
  }
}

void Grad::fddz(std::vector<float> &dpz,int adj) {
  if(adj==1) {
    #pragma omp parallel for
    for(int jgx=buff; jgx<nkgx+buff; jgx++) {
      #pragma omp simd
      for(int jz=buff; jz<nkz+buff; jz++) {
        int jj=jz+jgx*nkzEx;
        int ii=jz-buff+(jgx-buff)*nkz;
        dpz[ii] =
          b1z*(-dpz_cur[jj  ]+dpz_cur[jj-1])+
          b2z*(-dpz_cur[jj+1]+dpz_cur[jj-2])+
          b3z*(-dpz_cur[jj+2]+dpz_cur[jj-3])+
          b4z*(-dpz_cur[jj+3]+dpz_cur[jj-4])+
          b5z*(-dpz_cur[jj+4]+dpz_cur[jj-5])+
          b6z*(-dpz_cur[jj+5]+dpz_cur[jj-6]);
      }
    }
  }
  else {
    #pragma omp parallel for
    for(int jgx=buff; jgx<nkgx+buff; jgx++) {
      #pragma omp simd
      for(int jz=buff; jz<nkz+buff; jz++) {
        int jj=jz+jgx*nkzEx;
        int ii=jz-buff+(jgx-buff)*nkz;
        dpz[ii] =
          b1z*(dpz_cur[jj+1]-dpz_cur[jj  ])+
          b2z*(dpz_cur[jj+2]-dpz_cur[jj-1])+
          b3z*(dpz_cur[jj+3]-dpz_cur[jj-2])+
          b4z*(dpz_cur[jj+4]-dpz_cur[jj-3])+
          b5z*(dpz_cur[jj+5]-dpz_cur[jj-4])+
          b6z*(dpz_cur[jj+6]-dpz_cur[jj-5]);
      }
    }
  }
}
