#include "read.h"
#include <vector>
#include "wave_prop.h"
#include "scale.h"
#include "taper.h"
#include "snap.h"
#include <cmath>
#include "alloc.h"
#include "allocv.h"
#include "nlmod.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <iostream>
#include "randomBound.h"

void NLMod::nlmod_init(Read &par, WaveProp& fwd_in) {
  // Shorcuts for frequently object's members invokation
  pad=par.pad;
  zpad=par.zpad;
  nkgx=par.nkgx;
  nkz=par.nkz;
  nx=par.nx;
  nz=par.nz;
  nxE=par.nxE;
  nzE=par.nzE;
  nt=par.nt;
  napL=par.napL;
  napR=par.napR;
  dt=par.dt;
  dx=par.dx;
  use_taper=par.use_taper;
  datumS=par.datumS;
  datumR=par.datumR;
  fwd=&fwd_in; 
  dipole=par.dipole;

  // Allocate arrays for scattered wavefield computations
  allocV(dsynt,nkgx*nt);
  alloc(Vtop,nkgx);
  allocV(paux,nkgx*nkz);

  // Vertical position of the source (dipole source)
  jsrc1=           datumS  +pad+zpad+(pad+napL+par.shot_pos)*nkz;
  if(dipole) jsrc2=datumS+1+pad+zpad+(pad+napL+par.shot_pos)*nkz;
}

void NLMod::nlmod(Read &par,int direct_waves) {
  // If modeling direct waves only
  if(direct_waves==1) {
    // Prepare velocity at the surface for direct waves modeling
    for(int jgx=0; jgx<nkgx; jgx++) {
      int jj=zpad+pad+jgx*nkz; 
      Vtop[jgx]=(*fwd).Vshot[jj];
    }
    for(int jgx=0; jgx<par.nkgx; jgx++) {
      for(int jz=0; jz<par.nkz; jz++) {
        int jj=jz+jgx*nkz; 
        (*fwd).Vshot[jj]=Vtop[jgx];
      }
    }
  }

  // Initialize/update wavefield arrays
  (*fwd).init_wavefields();

  // Initialize scalar
  scale.scale_init(par,(*fwd).Vel2,0);

  // Initialize synthetic data
  if(par.adj == 1) {
    // Initialize synthetic source
    for(int jj=0; jj<par.src.size(); jj++) {
      par.src[jj] = 0.0;
    }
  }
  else {
    for(int jj=0; jj<dsynt.size(); jj++) {
      dsynt[jj] = 0.0;
    }
  }

  if(par.adj == 0) {
    std::cout << "Perform non-linear modeling (synthesize data)" << std::endl;
    for(int jt=0; jt<nt; jt++) {
      // Forward propagation of source wavefield
      if(jt>1) (*fwd).prop(0);
  
      if(jt>0) {
        // Prepare dipole source wavelet
        paux[jsrc1]           = par.src[jt-1];
        if(dipole) paux[jsrc2]=-par.src[jt-1];

        // Scale, then inject the source
        scale.scale(0,1,paux,(*fwd).p);

      // Damping function applied to the source wavefield
      if(par.use_taper == 1) 
        taper((*fwd).pm2,(*fwd).pm1,(*fwd).p);
      }
  
      // Save snapshots: source wavefield
      if(direct_waves == 0) snapshot(par,jt,(*fwd).p,1,"snap1");
  
      // Extract the wavefield at the receiver position
      #pragma omp parallel for
      for(int jgx=0; jgx<nkgx; jgx++) {
        int j1=datumR  +pad+zpad+jgx*nkz;
        int j2=datumR+1+pad+zpad+jgx*nkz;
        int j3=jt+jgx*nt;
        dsynt[j3]           +=(*fwd).p[j1];
        if(dipole) dsynt[j3]-=(*fwd).p[j2];
      }
    }
  }
  else {
    std::cout << "Perform non-linear adjoint (estimate the wavelet)" << std::endl;
    for(int jt=nt-1; jt>=0; jt--) {
      // Backward propagation of receiver wavefield
      if(jt<nt-2) (*fwd).prop(1);
  
      if(jt<nt-1) {
        // Prepare dipole data
        #pragma omp parallel for
        for(int jgx=0; jgx<nkgx; jgx++) {
          int j1=datumR  +pad+zpad+jgx*nkz;
          int j2=datumR+1+pad+zpad+jgx*nkz;
          int j3=jt+1+jgx*nt;
          (*fwd).p[j1]           +=par.rcv[j3];
          if(dipole) (*fwd).p[j2]-=par.rcv[j3];
        }
      }

      // Damping function applied to the source wavefield
      if(par.use_taper == 1) 
        taper((*fwd).pm2,(*fwd).pm1,(*fwd).p);
    
      // Scale the scattered source wavefield before propagation
      scale.scale(1,0,paux,(*fwd).p);

      // Save snapshots: source wavefield
      snapshot(par,jt,paux,1,"snap1");
  
      // Extract the wavefield at the source position
      par.src[jt]           +=paux[jsrc1];
      if(dipole) par.src[jt]-=paux[jsrc2];
    }
  }
}
