/*
  **************************************************************
  *** Main program for 2D nonlinear modeling using FD        ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <vector>
#include "wave_prop.h"
#include "scale.h"
#include "alloc.h"
#include "allocv.h"
#include <cmath>
#include "nlmod.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {
  // Define class member objects
  Read par;
  WaveProp fwd;
  NLMod nlmod; 
  Write out;
 
  std::cout << "**************************************************************" << std::endl;
  std::cout << "*** 2D nonlinear modeling using finite differences           *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                               *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017   *" << std::endl;
  std::cout << "**************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the input history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Build geomery from input parameters
   par.build_geometry();

  // Read velocity
  par.read_velocity();
 
  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();

  // Proceed with the following if dataset is provided
//  if(par.ntracesData != 0 && par.adj == 1) {
//    par.data_geometry();
//  }

  // Read and verbose source
  par.read_src(0);

  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize tapering
  taper_init(par);

  // Initialize some variables for the main loop
  par.ntSnap=0;

  // Allocate auxiliary arrays
  std::vector<float> d(par.nt*par.nkgx);

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Initialize parameters for forward propagation
  fwd.prop_init(par);

  // Create velocity frames with random boundaries to migrate shots
  par.vel_frames(fwd.Vshot,0);

  if(par.adj==1) {
    // Read input data for source estimation
    par.read_shot(0);
  }

  // Initialize parameters for non-linear modeling
  nlmod.nlmod_init(par,fwd);

  // Perform non-linear modeling for each source
  nlmod.nlmod(par,0);

  if(par.adj == 0) { // For data modeling only
    // Temporary stores modeled data
    for(int jj=0; jj<d.size(); jj++) {
        d[jj]=nlmod.dsynt[jj];
    }
  
    // Re-initialize synthetic data
    for(int jj=0; jj<nlmod.dsynt.size(); jj++) {
      nlmod.dsynt[jj] = 0.0;
    }
  
    // Perform non-linear modeling for direct waves
    nlmod.nlmod(par,1);
  
    // Subtract direct waves from non-linear modeled data
    for(int jj=0; jj<d.size(); jj++) {
      nlmod.dsynt[jj]=d[jj]-nlmod.dsynt[jj];
    }
  }
  std::cout << "*********** MAIN CALCULATION ENDS! ********** " << std::endl;

  // Initialize writing
  out.write_init(par);

  // Write the output
  if(par.adj == 0) {
    std::cout << "Write the output binary of modeled data" << std::endl;  
    out.write_data(par,nlmod.dsynt,"nldata");

    // Write the headers
//    out.write_data_headers(par,"nldata");
  }
  else {
    std::cout << "Write the output binary of modeled source" << std::endl;  
    out.write_data(par,par.src,"nlsrc");
  }

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "nlmod2d: Finishing successfully!!!" << std::endl;  

  hclose();

  return 0;

}
