/*
  **************************************************************
  *** Main program for 2D nonlinear modeling.                ***
  *** Using finite differences with Lax-Wendroff method      ***
  *** Version rewritten in C++                               ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <vector>
#include "lap.h"
#include "wave_prop.h"
#include "scale.h"
#include "vshot.h"
#include "alloc.h"
#include "allocv.h"
#include <cmath>
#include "nlmod.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <cstdlib>
void generateRicker(float * source, float fund, float dt, int nt, float tdelay) {
    for (int it = 0; it < nt; it++) {
    	double t = dt * it - tdelay;
        double tmp0 = PI * PI * fund * fund * t * t;
        source[it] = (1.0 - 2 * tmp0) * exp(-tmp0);
    }
}

using namespace std;
int MAIN() {
  // Define class member objects
  Read par;
  NLMod nlmod; 
  Write out;
 
  std::cout << "**************************************************************" << std::endl;
  std::cout << "*** 2D nonlinear modeling using FD with Lax-Wendroff method  *" << std::endl;
  std::cout << "*** Version rewritten in C++                                 *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                               *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017   *" << std::endl;
  std::cout << "**************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
//  par.read_parameters();

  // Read data history
//  par.read_history();
  
//  if(par.ntracesData==0) {// Read from parameters file
//     par.build_geometry();
//  }
//  else {                  // Read data header
//    par.read_headers();
//  }

  // Read velocity
//  par.read_velocity();
 
 
 
  // Determine time sampling according to Courant-Fredrich-Lewis condition
//  par.courant();

  // Proceed with the following if dataset is provided
//  if(par.ntracesData != 0 && par.adj == 1) {
//    par.data_geometry();

//    if(sep_copy_hff("datafile","nldata"))
//      seperr("nonlin2d: Cannot copy headers from datafile to nldatafile");
//  }

  // Read and verbose source
//  par.read_src(0);

// SETTING THE PARAMETERS BY HAND
  par.hmax=0;
  par.pad=0;
  par.zpad=0;

  par.nh=0;
  par.nx=401;
  par.ngx=401;
  par.nkgx=401;
  par.nbox=401;
  par.nz=401;
  par.ntraces=401;
  par.nkz=401;
  par.ntData=1000;
  par.nt=1000;

  par.dh=1;
  par.dx=10;
  par.dgx=10;
  par.dgxOdx=1;
  par.dz=10;
  par.dt=0.002;
  par.dtData=0.002;

  par.src_pad=0;
  par.shot_pos=200;
//  par.cur_shot=1;

  par.nthreads=1;
  par.first_cdp=1;
  par.last_cdp=par.nx;
  par.apertureL=0.;
  par.apertureR=0.;
  par.napL=0;
  par.napR=0;
  par.napR=0;
  par.trace_shot=0;
  par.jtSnap=1;
  par.adj=0;

  par.Vmin=2000;
  par.Vmax=2000;
  par.datumS=floor(par.nx/2);
  par.datumR=floor(par.nx/2);

  alloc(par.sx,1);
  alloc(par.gx,1);
  par.sx[0]=0.5*(par.nx-1)*par.dx;
  par.gx[0]=0;

  // velocity  
  alloc(par.Vint, par.nx, par.nz);
  float **val = par.Vint;

  for (int i = 0; i < par.nx; i++) {
  	for (int j = 0; j < par.nz; j++) {
  		par.Vint[i][j] = 2000.;
  	}
  }
  // source wavelet
  float fund = 10., dt = 0.002, tdelay = 0.2;
  float *source = new float[par.nt];
  generateRicker(source, fund, dt, par.nt, tdelay);
  for (int i = 0; i < par.nt; i++) {
  	par.src.push_back(source[i]);
  }	

 // Geometric array
 alloc(par.sgx,par.nkgx);
 for(int jgx=0; jgx<par.nkgx; jgx++) {
   par.sgx[jgx] = jgx*par.dgx;
 }

  // Verbose some useful parameters
  par.verbose_par();
  int nx = par.nx, nz = par.nz, nt = par.nt;
//  float *fake = new float[nx * nz * nt];
//  for (int i = 0; i < nx * nz * nt; i++) {
//  	fake[i] = 0.1 * rand() / RAND_MAX; 
//  }
//  srite("snap1", fake, sizeof(float) * nx * nz * nt);
//  return 0;
  // Initialize tapering
//  Read &ppar=par; // Reference to make changes in "par" members within a function 
//  taper_init(ppar);

  // Initialize parameters for non-linear modeling
  nlmod.nlmod_init(par);

  // Initialize some variables for the main loop
  par.ntSnap=0;
  par.first_geof=0;

  // Allocate auxiliary arrays
  std::vector<float> d(par.nt*par.nkgx);

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Get number of geophone traces in shot gather (in some datasets is variable)
  par.ngeof = 0; //par.source_index[0];
  par.last_geof = par.first_geof+par.ngeof;
  cout << "am I here >?" << endl;

  // Perform non-linear modeling for each source
  nlmod.nlmod(par,0);

  std::cout << "*********** MAIN CALCULATION ENDS! ********** " << std::endl;

  // Initialize writing
  out.write_init(par);

  // Write the output
  out.write_snap("snap1");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "nlmod2d: Finishing successfully!!!" << std::endl;  

  hclose();

  return 0;

}
