/*
  **************************************************************
  *** Main program for 2D Hessian computation with PSF.      ***
  *** Applies Born mod. followed by RTM on seeded spikes.    ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "lint2.h"
#include "base_trismt.h"
#include <cmath>
#include "alloc.h"
#include "born.h"
#include "rtm.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {
  // Define class member objects
  Read par;
  WaveProp fwd, scat, bwd;
  Born born; 
  Rtm rtm; 
  Write out;

  std::cout << "******************************************************************" << std::endl;
  std::cout << "*** 2D Hessian computation using point-spread functions (PSF)    *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                   *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017       *" << std::endl;
  std::cout << "******************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the input history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Build geomery from input parameters
  par.build_geometry();

  // Read velocity
  par.read_velocity();

  // Read reflectivity (actually creates zero-filled array)
  par.read_reflectivity();

  // Overwrite reflectivity with spikes
  int npsfx, npsfz;
    npsfx=0;
    for(int jx=par.incPSFx; jx<par.nxE; jx+=par.incPSFx) {
      npsfz=0;
      for(int jz=par.incPSFz; jz<par.nzE; jz+=par.incPSFz) {
        int jj=jz+jx*par.nzE+par.nh*par.nxE*par.nzE;
        if(jz > par.zpad+par.pad && jz < par.nz+par.zpad+par.pad) {
          par.refl[jj] = 1.0;
        }
        npsfz++;
      }
      npsfx++;
    }

  std::cout << "Point-spread functions in X: npsfx = " << npsfx << std::endl;
  std::cout << "Point-spread functions in Z: npsfz = " << npsfz << std::endl;
  par.npsfx=npsfx;
  par.npsfz=npsfz;

  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Read and verbose source
  par.read_src(1);
  
  // Verbose some useful parameters
  par.verbose_par();
  
  // Initialize tapering for receiver propagation
  taper_init(par);

  // Initialize some variables for the main loop
  par.ntSnap=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Initialize the output
  out.write_init(par);

  // Initialize parameters for source and scattered wavefields
  fwd.prop_init(par);
  bwd.prop_init(par);
  scat.prop_init(par);
  
  // Create velocity frames
  par.vel_frames(fwd.Vshot,par.rbc);
  par.vel_frames(bwd.Vshot,par.rbc);
  par.vel_frames(scat.Vshot,par.rbc);
  
  std::cout << "Obtain psf from Born/RTM pair" << std::endl;

  // Initialize parameters for Born modeling
  born.born_init(par,fwd,scat);
  
  // Perform born modeling
  born.born(par);

  // Pass the Born outcome as input for RTM
  par.rcv.swap(born.dsynt);

  // Re-create velocity frames for source wavefield
  par.vel_frames(fwd.Vshot,1);
  
  // Initialize parameters for RTM
  rtm.rtm_init(par,fwd,bwd);
  
  // Perform Reverse-Time Migration
  rtm.rtm(par);

  // Construct the images
  rtm.build_images(par);

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;

  // Write the PSFs in file
  std::cout << "Write the Born/RTM PSFs" << std::endl;  

  // Write the whole volume with PSFs 
  out.write_psf(par,rtm.ICorr.data());

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count()
                                << "secs." << std::endl; 
  std::cout << "psf2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}
