/*
  ********************************************************************
  *** This class encompasses functions for reading different input ***
  *** parameters, both from command line and files.                ***
  ********************************************************************
*/
#include "interp.h"
#include "read.h"
#include "minmax.h"
#include "alloc.h"
#include "allocv.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <iostream>
#include <vector>
#include <string>
#include <cmath>
#include <omp.h>

void Read::read_parameters() {
  /* Read parameters from command line */
  std::cout << "read_parameters: Reading parameters from command line" << std::endl;

  // Factor applied to velocity model
  if(0 == getch("vel_frac","f",&vel_frac))
    vel_frac=1.0;

  // Input file for source wavefield
  if(0 == getch("wavelet_fn","s",&char_aux)) {
    std::cout << "read_parameters: Warning: Input wavelet missing!" << std::endl;
  }
  wavelet_fn=char_aux;

  // Input file for reflectivity
  if(0 == getch("reflfile","s",&char_aux))
    reflfile="";
  else
    reflfile=char_aux; 

  // Input file for field data
  if(0 == getch("datafile","s",&char_aux)) {
    getch("projpref","s",&projpref);
  }
  else {
    projpref="none";
   datafile=char_aux;
  } 

  // Trace within the shot where shot
  if(0 == getch("trace_shot","d",&trace_shot)) 
    trace_shot=1; 

  // No. of point for triangle filter (deconvoution imaging cond.)
  if(0 == getch("npts","d",&npts)) 
    npts=0; 

  // Override header's scalco
  if(0 == getch("scalco_override","d",&scalco)) 
    scalco=0;

  // Factor to time
  if(0 == getch("time_frac","f",&time_frac)) 
    time_frac=1.0; 

  // Set not to use the taper (it activates if taper is initialized)
  if(0 == getch("use_taper","d",&use_taper)) 
  use_taper=0;

  // "Aggressiveness" of the random BC
  if(0 == getch("rand_coeff","f",&rand_coeff))
    rand_coeff=1.;

  // Number of threads if using openmp
  if(0 == getch("nthreads","d",&nthreads))
    nthreads=1; 

  // Set the number of threads for openmp
    omp_set_num_threads(nthreads); 

  // Number of samples for the input wavefield
  if(0 == getch("nts","d",&ntsData))
    ntsData=ntData;

  // Time delay for the input source wavelet
  if(0 == getch("src_delay","d",&src_delay))
    src_delay=0;

  // Want dipole source?
  if(0 == getch("dipole","d",&dipole))
    dipole=1;
  
  // Verbose the source values?
  if(0 == getch("verb_src","d",&verb_src))
    verb_src=0;  

  // Use DSO?
  if(0 == getch("dso","d",&dso)) 
    dso=0; 

  if(0 == getch("neg_dso","d",&neg_dso)) 
    neg_dso=0; 

  // Extend in subsurface offset?
  if(0 == getch("extend","d",&extend)) 
    extend=0;
  if(dso == 1) extend=1;

  // Parameters for offset-domain CIG
  if(0 == getch("dh","d",&dh))
    dh=1;
  if(0 == getch("hmax","d",&hmax))
    hmax=0;
  if(extend == 0) hmax=0;
  if(remainder(hmax,dh) == 0)
    nh=hmax/dh;
  else{
    seperr("read_parameters: nh is not an integer, check hmax and dh");
  }

  // Factors to coarsen/refining the grid for B-spline interpolation
  if(0 == getch("factX","f",&factX)) 
    factX=1; 

  if(0 == getch("factZ","f",&factZ)) 
    factZ=1; 

  // Parameters of the energy imaging condition
  if(0 == getch("eic","d",&eic))
    eic=0;

  // Shift for the X coordinate to make them divisible by the sampling rate
  // Parameters for computing source snapshots
  if(0 == getch("jtSnap","d",&jtSnap))
    jtSnap=1;

  // Datum of the source wavelet
  if(0 == getch("datumS","d",&datumS))
    datumS=0;

  // Datum of the receiver data
  if(0 == getch("datumR","d",&datumR))
    datumR=0;

  // Compute PSF in Born/RTM or Wemva/Wemva?
  if(0 == getch("psf_born","d",&psf_born))
    psf_born=0;

  // Number of lags for Hessian computation
  if(0 == getch("nlagz","d",&nlagz))
    nlagz=0;
  if(0 == getch("nlagx","d",&nlagx))
    nlagx=0;
  if(0 == getch("incPSFz","d",&incPSFz))
    incPSFz=1;
  if(0 == getch("incPSFx","d",&incPSFx))
    incPSFx=1;

  // For nonlinear modeling only: Keep sources or recalculate
  if(0 == getch("keep_shots","d",&keep_shots))
    keep_shots=0;

  // For nonlinear modeling only: Shot number to be modeled
  // Employed too when no header file is provided
  if(0 == getch("cur_shot","d",&cur_shot))
    cur_shot=1;

  // When needed, one specifies forward or adjoint
  if(0 == getch("adj","d",&adj))
    adj=0;

  // Use or not random boundary conditions RBC
  if(0 == getch("rbc","d",&rbc))
    rbc=1;

  // For dot product test
  if(0 == getch("dot_prod","d",&dot_prod))
    dot_prod=0;
  if(dot_prod==1) {
    rbc=0;
    use_taper=0;
  }
}

void Read::read_history() {
  std::cout << "read_history: Reading parameters from history files" << std::endl;

  // Parameters from history files
  int n1, n2;
  if(0 == hetch("n1","d",&n1))
    seperr("read_history: Parameter nz not found\n"); 
  if(0 == hetch("n2","d",&n2))
    seperr("read_history: Parameter nx not found\n"); 
  if(0 == hetch("o1","f",&o1))  o1=0.0;
  if(0 == hetch("o2","f",&o2))  o2=0.0;
  if(0 == hetch("d1","f",&dz))
    seperr("read_history: Parameter d1 not found\n"); 
  if(0 == hetch("d2","f",&d2))
    seperr("read_history: Parameter d2 not found\n"); 
  dx=d2;

  // Migration apertures (in grdpoints)
  if(0 == hetch("napL","d",&napL)) 
    seperr("read_history: Parameter napL not found\n"); 
  if(0 == hetch("napR","d",&napR)) 
    seperr("read_history: Parameter naplR not found\n"); 

  // Padding (in gridpoints)
  if(0 == hetch("pad","d",&pad)) 
    seperr("read_history: Parameter pad not found\n"); 
  if(0 == hetch("zpad","d",&zpad)) 
    seperr("read_history: Parameter pad not found\n"); 

//  if(0 == hetch("unit1","s",&unit1))  unit1="";
//  if(0 == hetch("unit2","s",&unit2))  unit2="";

  // Parameters read from auxiliary files
  if(!datafile.empty()) { // If data geometry is copied
    if(0 == auxpar("n1","d",&ntData,"datafile"))
      seperr("read_history: Parameter ntData not found\n"); 
    if(0 == auxpar("n2","d",&ntraces,"datafile"))
      seperr("read_history: Parameter ntraces not found\n"); 
    if(0 == auxpar("d1","f",&dtData,"datafile"))
      seperr("read_history: Parameter dtData not found\n"); 
  }
  // Define some parameters
  nz=n1-zpad-2*pad;
  nx=n2-2*pad-napL-napR;
  nzE=n1;
  nxE=n2;
  o1=o1+(zpad+pad)*dz;
  o2=o2+(pad+napL)*dx;
  first_cdp = 1+floor(o2/d2);
  last_cdp = first_cdp+(nx-1);
  t_end = (ntData-1)*dtData;
}

void Read::read_headers() {
  std::cout << "read_headers: Reading parameters from header files" << std::endl;
  int error=0;
  // Read keys from header file
  if(0 != sep_get_number_keys("datafile",&n_keys_data)) {
    std::cout << "Cannot read in number of keys from the hff (.H@@) file!" << std::endl;
    error++;
  }
  if(0 != sep_get_key_index("datafile","sx",&sx_key)) {
    std::cout << "sx key index does not exist on datafile file!" << std::endl;
    error++;
  }
  if(0 != sep_get_key_index("datafile","gx",&gx_key)) {
    std::cout << "gx key index does not exist on datafile file!" << std::endl;
    error++;
  }
  if(0 != sep_get_key_index("datafile","sy",&sy_key)) {
    std::cout << "sy key index does not exist on datafile file!" << std::endl;
    error++;
  }
  if(0 != sep_get_key_index("datafile","gy",&gy_key)) {
    std::cout << "gy key index does not exist on datafile file!" << std::endl;
    error++;
  }
  if(0 != sep_get_key_index("datafile","scalco",&scalco_key)) {
    std::cout << "scalco key index does not exist on datafile file!" << std::endl;
    error++;
  }

  if(error == 6) {
    std::cout << "There is no header file!" << std::endl;
    std::cout << "Will try to create the geometry from the parameters instead" << std::endl;
    build_geometry();
    return;
  }

  // Allocate header handler array
  alloc(header_in,ntraces,n_keys_data);

  for(int jj=0; jj<ntraces; jj++) {
    for(int ii=0; ii<n_keys_data; ii++) {
      header_in[jj][ii]=0.0;
    }
  }
    
  // Read geometry from header file 
  int trace=1;
  for(int jj=0; jj<ntraces; jj++) { 
    previous_trace=jj+1;
    if(0 != sep_get_val_headers("datafile",&previous_trace,&trace,header_in[jj]))
      seperr("could not read from input data headers!");
  }

  int nshift=pad+napL;
  // Obtain scalco
  if(scalco == 0) scalco=header_in[0+nshift][scalco_key-1];
  scalcof=scalco;
  if(scalcof < 0) scalcof=-1.0/scalcof;

  // Define source and receiver spacings
  dsx = scalcof*(header_in[1+nshift][sx_key-1]-header_in[0+nshift][sx_key-1]); // This might be obsolete
  dgx = scalcof*(header_in[1+nshift][gx_key-1]-header_in[0+nshift][gx_key-1]);

  // Factorize the traces by dgx/dx ratio 
  if(dgx != dx)
    seperr("read_headers: dgx and dx are different!");

  // Allocate geometry arrays
  sx = new float[ntraces];
  gx = new float[ntraces];
  hoff = new float[ntraces];
  source_index = new int [2];

  // Prepare source and receiver geometries
  for(int jtraces=0; jtraces<ntraces; jtraces++) {
    sx[jtraces]=header_in[jtraces][sx_key-1]*scalcof;
    gx[jtraces]=header_in[jtraces][gx_key-1]*scalcof;
  }

  // Offset calculation
  for(int jtraces=0; jtraces<ntraces; jtraces++) {
    hoff[jtraces]=fabs(gx[jtraces]-sx[jtraces]);
  }
  hmin=minval(hoff,ntraces);
  src_pad=floor(hmin/dx);

  // Deallocate header array
  delete header_in;
}

void Read::read_velocity() {
  std::cout << "read_velocity: Begin reading velocities" << std::endl;

  // Alocate velocity array
  int n=nxE*nzE;
  allocV(Vint,n);

  // Read velocity from binary
  if(n*4 != sreed("in",Vint.data(),n*4))
    seperr("read_velocity: Cannot read velocity file");

  std::cout << "read_velocity: Finish reading velocities" << std::endl;

  // Obtain some useful quantities: mean, minimum and maximum vels.
  float sum=0.0;
  for(int jj=0; jj<Vint.size(); jj++) {
    sum+=Vint[jj];
  }
  Vcte=sum/(n);
  Vmin=minvalV(Vint,n);
  Vmax=maxvalV(Vint,n);
}

void Read::read_reflectivity() {

  if(0 == auxin("reflfile")) {
    // Read reflectivity from binary
    std::cout << "read_reflectivity: Allocate reflectivity vector only" << std::endl;

    // Allocate reflectivity array
    int n=nzE*nxE*(2*nh+1);
    allocV(refl,n);
  }
  else {
    // Override offset value
    int naux;
    if(0 == auxpar("n3","d",&naux,"reflfile"))
      seperr("read_reflectivity: Cannot read n3 in reflfile \n"); 
    nh=(naux-1)/2;
    hmax=nh;
    if(hmax > 0) extend=1;

    // Allocate reflectivity array
    int n=nzE*nxE*(2*nh+1);
    allocV(refl,n);

    std::cout << "read_reflectivity: Begin reading reflectivity" << std::endl;
    // Read reflectivity from binary
    if(n*4 != sreed("reflfile",refl.data(),n*4))
      std::cout << "read_reflectivity: WARNING: Cannot read reflectivity file" << std::endl;
    std::cout << "read_reflectivity: Finish reading reflectivity" << std::endl;
     auxclose("reflfile");
  }
}

void Read::read_refl_hess(int nd) {
  
  int n1;
  if(0 == auxpar("n1","d",&n1,"reflfile"))
    seperr("read_refl_hess: Parameter n1 not found\n"); 
  int n2;
  if(0 == auxpar("n2","d",&n2,"reflfile"))
    seperr("read_refl_hess: Parameter n2 not found\n"); 
  int n3;
  if(0 == auxpar("n3","d",&n3,"reflfile"))
    seperr("read_refl_hess: Parameter n3 not found\n"); 
  int nzxh=n1*n2*n3;

  if(adj == 1 && nd==nzxh) {
    std::cout << "read_refl_hess: Begin reading reflectivity" << std::endl;
    // Allocate reflectivity array
    int n=(nxE+2*nlagx)*(nzE+2*nlagz)*(2*nh+1);
    allocV(refl,n);
  
    // Read reflectivity from binary
    if(n*4 != sreed("reflfile",refl.data(),n*4))
      seperr("read_reflectivity: Cannot read reflectivity file");
    auxclose("reflfile");
    std::cout << "read_refl_hess: Finish reading reflectivity" << std::endl;
  }
  else {
    std::cout << "read_refl_hess: Re-use read_reflectivity\n" << std::endl;
    read_reflectivity();
  }
}

void Read::read_pert(int adj) {
  if(adj==1) { // Re-use read_reflectivity function
    std::cout << "read_pert: Begin reading perturbation in reflectivity" << std::endl;
    read_reflectivity();
    std::cout << "read_pert: Finish reading perturbation in reflectivity" << std::endl;
  }
  else { // Slowness has no extension, thus needs different treatment
    std::cout << "read_pert: Begin reading perturbation in slowness" << std::endl;

    // Allocate reflectivity array
    int n=nxE*nzE*(2*nh+1);
    allocV(refl,n);

    // Read pert. in slowness from binary
    std::vector<float> paux(nzE);
    for(int jx=0; jx<nxE; jx++) {
      if(nzE*4 != sreed("reflfile",paux.data(),nzE*4))
        seperr("read_pert: Cannot read pert. in slowness file");
      for(int jz=0; jz<nzE; jz++) {
        int jj=jz+jx*nzE+nh*nxE*nzE;
        refl[jj] = paux[jz];
      }
    }
    std::cout << "read_pert: Finish reading perturbation in slowness" << std::endl;
    auxclose("reflfile");
  }
}

void Read::vel_frames(std::vector<float>& Vshot,int read) {
  // This function either creates or reads velocity frames for the
  // corresponding shot gather (the latter for RBC)
  if(read ==1) {
    // Read frames
    int n=nkgx*nkz;
    if(n*4 != sreed("vel_frame",Vshot.data(),n*4))
      seperr("vel_frames: Cannot read vel_frame file");
    auxclose("vel_frame");
  }
  else {
    if(nkgx==nxE) {
      Vshot=Vint;
    }
    else {
      int jx;
      for(int jgx=0; jgx<nkgx; jgx++) {
        float coord=sgx[jgx]+(pad+napL)*dx;
        if(coord>=0 && coord<nxE*dx) {
          jx=floor(coord/dx); //+pad+napL;
        }
        else if(coord<0) {
          jx=0;
        }
        else { 
          jx=nxE-1;
        }
        for(int jz=0; jz<nkz; jz++) {
          int j1=jz+jgx*nkz;
          int j2=jz+jx*nzE;
          Vshot[j1]=Vint[j2];
        }
      }
    }
  }
}

void Read::refl_frames(std::vector<float>& rshot) {
  // This function creates reflectivity frames for the
  // corresponding shot gather
  if(nkgx==nxE) {
    rshot.swap(refl);
  }
  else {
    int jx;
    for(int jh=0; jh<2*nh+1; jh++) {
      for(int jgx=0; jgx<nkgx; jgx++) {
        float coord=sgx[jgx]+(pad+napL)*dx;
        if(coord>=0 && coord<nxE*dx) {
          jx=floor(coord/dx); //+pad+napL;
        }
        else if(coord<0) {
          jx=0;
        }
        else { 
          jx=nxE-1;
        }
        for(int jz=0; jz<nkz; jz++) {
          int j1=jz+jgx*nkz+jh*nkgx*nkz;
          int j2=jz+jx*nzE+jh*nxE*nzE;
          rshot[j1]=refl[j2];
        }
      }
    }
  }
  
}

void Read::courant() {
  // Obtain the Courant number
  if(dx <= dz)
    cfl = Vmax*dtData/dx;
  else
    cfl = Vmax*dtData/dz;
  tIntFact = ceil(cfl/0.7);
  dt = dtData/tIntFact;
  nt = 1+round(t_end/dt);
}

void Read::read_src(int der2) {
  std::cout << "Read source from file" << std::endl;

  // Allocate source arrays
  std::vector<float> srcData(ntData,0.);
  if(src.empty()) allocV(src,nt);

  // Allocate receiver arrays
  if(rcv.empty()) allocV(rcv,nkgx*nt);

  // Read source from binary
  if(ntData*4 != sreed("wavelet_fn",srcData.data(),ntData*4))
    seperr("read_src: Cannot read source file");
  
  // Verbose the source
  for(int jj=0; jj<ntData; jj++) {
    if(verb_src == 1) {
      std::cout << "srcData: " << srcData[jj] << std::endl;
    }
  }
  
  // Interpolate time samples to satisfy Courant condition
  int doit=0;
  if(ntData != nt) doit=1;
  Interp interp(tIntFact,doit);
  if(ntData != nt) {
    interp.op(0,0,srcData,src);
  }
  else {
    tIntFact=1;
    for(int jt=0; jt<nt; jt++)
      src[jt]=srcData[jt];
  }
  nts=tIntFact*ntsData;

  if(der2 == 1) {
    // Use the second order time derivative of the source
    std::vector<float> srcTemp(nt,0.);
    srcTemp[src_delay] = src[0]/dt/dt;
    srcTemp[src_delay+1] = (src[1]-2*src[0])/dt/dt;
    for(int jt=src_delay+2; jt<nt; jt++) {
      srcTemp[jt] = (src[jt]-2*src[jt-1]+src[jt-2])/dt/dt; 
    }
    src.swap(srcTemp);
  }

  // Zero the elements beyond nts
  for(int jt=nts; jt<nt; jt++) src[jt]=0;
  
  // Obtain Nyquist frequency
  fNyq = 0.5/dtData;
  
  // Read the cut-off frequency (by default is Nyquist)
  if(0 == getch("fc","f",&fc)) fc=fNyq;

  // Set the step size at which scatter/imag. cond. will be applied
  scatStep=floor(0.5/(dt*fc));
  if(scatStep < 1) scatStep=1;

  auxclose("wavelet_fn");
}

void Read::read_shot(int der2) {
  std::cout << "Read shot file" << std::endl;

  std::vector<float> rcvData(ntData,0.);
  std::vector<float> temp(nt,0);
  int doit=0;
  if(psf_born==0) doit=1;
  Interp interp(tIntFact,doit);
  if(psf_born==0) {
    for(int jgx=0; jgx<nkgx; jgx++) {
      // Read data traces from binary
      if(4*ntData != sreed("datafile",rcvData.data(),4*ntData))
        seperr("read_shot: Have trouble reading trace");
  
      // Interpolate time samples to satisfy Courant condition
      if(tIntFact > 1) {
        interp.op(0,0,rcvData,temp);
      }
      else {
        temp=rcvData;
      }
      for(int jt=0; jt<nt; jt++) {
        int jj=jt+jgx*nt;
        rcv[jj]=temp[jt];
      } 
    } 
  }

  if(der2 == 1) {
    // Use the second order time derivative of the data
    std::vector<float> rcvTemp(rcv.size(),0.);
    for(int jgx=0; jgx<nkgx; jgx++) {
      int j1=0+jgx*nt;
      int j2=1+jgx*nt;
      rcvTemp[j1] = rcv[j1]/dt/dt;
      rcvTemp[j2] = (rcv[j2]-2*rcv[j1])/dt/dt;
      for(int jt=2; jt<nt; jt++) {
         int jj = jt+jgx*nt; 
         rcvTemp[jj] = (rcv[jj]-2*rcv[jj-1]+rcv[jj-2])/dt/dt; 
      }
    }
    rcv.swap(rcvTemp);
  }
  auxclose("datafile");
}

void Read::build_geometry() {
/*
 This function creates geometry for nonlinear modeled data,
 or when header files are not provided
*/
  std::cout << "Build geometry for nonlinear modeling" << std::endl;

  // Read parameters for geometry
  if(0 == getch("dsx","f",&dsx))
    seperr("par_geometry: Parameter dsx not found");
  if(0 == getch("osx","f",&osx))
    seperr("par_geometry: Parameter osx not found");
  if(0 == getch("ngx","d",&ngx))
    seperr("par_geometry: Parameter ngx not found");
  if(0 == getch("dgx","f",&dgx))
    seperr("par_geometry: Parameter dgx not found");
  if(0 == getch("ogx","f",&ogx))
    seperr("par_geometry: Parameter ogx not found");
  if(0 == getch("ntData","d",&ntData))
    seperr("par_geometry: Parameter ntData not found");
  if(0 == getch("dtData","f",&dtData))
    seperr("par_geometry: Parameter dtData not found");

  if(dgx != dx) 
    seperr("buid_geometry: dgx and dx are different!");

  int src_no=cur_shot-1;

  // Define apertures and diverse array "sizes"
  apertureL=napL*dx;
  apertureR=napR*dx;
  nbox = ngx+napL+napR;
  nxpad = nbox+pad+pad;
  nkgx = nxpad;
  nkz = nz+pad+pad+zpad;

  // Allocate geometry arrays
  ntraces=nkgx;
  sx = new float[ntraces];
  gx = new float[ntraces];
  hoff = new float[ntraces];
  source_index = new int[2];

  for(int jgx=0; jgx<nkgx; jgx++) {
    sx[jgx]=UNDEFINED;
    gx[jgx]=UNDEFINED;
  }

  // Prepare source and receiver geometries
  for(int jgx=pad+napL; jgx<pad+napL+ngx; jgx++) {
    int jgx_shift=jgx-pad-napL;
    sx[jgx]=osx+src_no*dsx; 
    gx[jgx]=osx+src_no*dsx+ogx+jgx_shift*dgx;
  }

  // Offset calculation
  for(int jtraces=0; jtraces<ntraces; jtraces++) {
    hoff[jtraces]=fabs(gx[jtraces]-sx[jtraces]);
  }
  hmin=minval(hoff,ntraces);
  src_pad=floor(hmin/dx);
  t_end=(ntData-1)*dtData;

  // Allocate geometry matrices
  sgx = new float[nkgx];
 
 // Prepare source and receiver geometries
  std::cout << "build_geometry: Prepare source and receiver geometries" << std::endl;
  for(int jgx=0; jgx<nkgx; jgx++)
    sgx[jgx] = UNDEFINED;

  // Create the geometry matrix
  geom_matrix();

  // Obtain the maximum values of the receiver coordinates
  sgx_max=maxval(sgx,nkgx);

  // Now extend the coordinates into the aperture and padding regions
  int jj=1;
  for(int jgx=pad+napL+ngx; jgx<pad+napL+ngx+napR+pad+src_pad; jgx++) {
    sgx[jgx] = sgx[pad+napL+ngx-1]+jj*dx;
    jj++;
  }
  for(int jgx=0; jgx<pad+napL; jgx++) {
    sgx[jgx] = sgx[pad+napL]-(pad+napL-jgx)*dx;
  }
  shot_pos=floor((sx[pad+napL]-gx[pad+napL])/dx);
}

void Read::data_geometry() {
  std::cout << "Build geometry for imaging" << std::endl;
  // Obtain the number of live traces in shot gather
  geom_numbers();


  // Define apertures and diverse array "sizes"
  apertureL=napL*dx;
  apertureR=napR*dx;
  nbox = ngx+napL+napR;
  nxpad = nbox+pad+pad;
  nkgx = nxpad+src_pad;
  nkz = nz+pad+pad+zpad;

  // Allocate geometry matrices
  sgx = new float[nkgx];
 
 // Prepare source and receiver geometries
  std::cout << "data_geometry: Prepare source and receiver geometries" << std::endl;
  for(int jgx=0; jgx<nkgx; jgx++) {
    sgx[jgx] = UNDEFINED;
  }

  // Create the geometry matrix
  geom_matrix();

  // Obtain the maximum values of the receiver coordinates
  sgx_max=maxval(sgx,nkgx);
  std::cout << "sgx_max: " << sgx_max  << std::endl;

  // Now extend the coordinates onto the aperture regions
  int jj=1;
  for(int jgx=pad+napL+ngx; jgx<pad+napL+ngx+napR+pad+src_pad; jgx++) {
    sgx[jgx] = sgx_max+jj*dx;
    jj++;
  }
  for(int jgx=0; jgx<pad+napL; jgx++) {
    sgx[jgx] = sgx[pad+napL]-(pad+napL-jgx)*dx;
  }

  // Obtain the minimum values of the receiver coordinates
  sgx_min=minval(sgx,nkgx);
  std::cout << "sgx_min: " << sgx_min  << std::endl;

  // Set the first and last source values
  shot_pos=floor((sx[pad+napL]-gx[pad+napL])/dx);
}

void Read::verbose_par() {
  std::cout << "***************************" << std::endl;
  std::cout << "* Verbose some parameters *" << std::endl;
  std::cout << "***************************" << std::endl;
  std::cout << "cur_shot " << cur_shot << std::endl;
  std::cout << "nz " << nz << std::endl;
  std::cout << "nzE " << nzE << std::endl;
  std::cout << "ntData " << ntData << std::endl;
  std::cout << "nt " << nt << std::endl;
  std::cout << "ntsData " << ntsData << std::endl;
  std::cout << "nts " << nts << std::endl;
  std::cout << "nx " << nx << std::endl;
  std::cout << "nxE " << nxE << std::endl;
  std::cout << "ngx " << ngx << std::endl;
  std::cout << "nkz " << nkz << std::endl;
  std::cout << "nkgx " << nkgx << std::endl;
  std::cout << "nbox " << nbox << std::endl;
  std::cout << "ntraces " << ntraces << std::endl;
  std::cout << "nh " << nh << std::endl;
  std::cout << "Points triang filt " << npts << std::endl;
  std::cout << "first_cdp " << first_cdp << std::endl;
  std::cout << "last_cdp " << last_cdp << std::endl;
  std::cout << "pad " <<  pad << std::endl;
  std::cout << "src_pad " << src_pad << std::endl;
  std::cout << "Global av. vel. " << Vcte << std::endl;
  std::cout << "Vel. min. " << Vmin << std::endl;
  std::cout << "Vel. max. " << Vmax << std::endl;
  std::cout << "Vel. frac. " << vel_frac << std::endl;
  std::cout << "Scattering time step (scatStep) " << scatStep << std::endl;
  std::cout << "Max. freq. in data " << fc << std::endl;
  std::cout << "dsx " << dsx << std::endl;
  std::cout << "dgx " << dgx << std::endl;
  std::cout << "dx " << dx << std::endl;
  std::cout << "dtData " << dtData << std::endl;
  std::cout << "dt " << dt << std::endl;
  std::cout << "dz " << dz << std::endl;
  std::cout << "dh " << dh << std::endl;
  std::cout << "hmax " << hmax << std::endl;
  std::cout << "Extend? " << extend << std::endl;
  std::cout << "Use DSO? " << dso << std::endl;
  std::cout << "B-spline factor in X " << factX << std::endl;
  std::cout << "B-spline factor in Z " << factZ << std::endl;
  std::cout << "Courant number " << cfl << std::endl;
  std::cout << "Aperture left " << apertureL << std::endl;
  std::cout << "Aperture right " << apertureR << std::endl;
  std::cout << "napL " << napL << std::endl;
  std::cout << "napR " << napR << std::endl;
  std::cout << "jtSnap " << jtSnap << std::endl;
  std::cout << "scalco " << scalco << std::endl;
  std::cout << "nthreads " << nthreads << std::endl;
  std::cout << "trace_shot " << trace_shot << std::endl;
  std::cout << "shot_pos " << shot_pos << std::endl;
  std::cout << "RBC " << rbc << std::endl;
  std::cout << "dot_prod " << dot_prod << std::endl;
  std::cout << "Adjoint (for Hess and WEMVA) " << adj << std::endl;
  std::cout << "lagsx (Hessian only) " << nlagx << std::endl;
  std::cout << "lagsz (Hessian only) " << nlagz << std::endl;
  std::cout << "incPSFx (Hessian only) " << incPSFx << std::endl;
  std::cout << "incPSFz (Hessian only) " << incPSFz << std::endl;
  std::cout << "use_taper " << use_taper << std::endl;
  std::cout << "psf_born " << psf_born << std::endl;
}

void Read::geom_numbers() {
  // This private function uses the geometry vectors sx and gx
  // (previously obtained from the headers) to calculate the
  // number of shots, nsources, and the number of traces by shot, ngx.
  int k=0;
  for(int jtraces=0; jtraces<ntraces; jtraces++) {
    if(sx[jtraces] != UNDEFINED) k++;
  }
  ngx=k;
}

void Read::geom_matrix() {
  // This private function merely creates a geometry matrix,
  // which is useful for mapping from data coordinates
  // to model coordinates
  int k=1;
  ngx_prev=1;
  for(int jgx=0; jgx<ngx; jgx++) {
    int jtraces=jgx+pad+napL;
    if(jtraces < ntraces-1) {
      if(sx[jtraces] == sx[jtraces+1]) {
        sgx[k-1+napL+pad] = gx[jtraces];
        k++;
      }
      else {
        if(k > ngx_prev) {
          ngx_cur=k;
          ngx_prev=k;
        }
        source_index[0] = k;
        source_index[1] = jtraces;
        sgx[k-1+napL+pad] = gx[jtraces];
        k=1;
      }
    }
    else {
      sgx[k-1+napL+pad] = gx[jtraces];
      source_index[0] = k-1;
      source_index[1] = jtraces;
     
      if(k < ngx_cur)  k++;
      else {
        if(k > ngx_prev) {
          ngx_cur = k;
          ngx_prev = k;
        }
        k=1;
      } 
    }
  }
}

