/*
  **************************************************************
  *** Main program for 2D reverse-time migration.            ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "interp1d.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "alloc.h"
#include "allocv.h"
#include <vector>
#include <cmath>
#include "rtm.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {
  // Define class member objects
  Read par;
  WaveProp fwd, bwd;
  Rtm rtm; 
  Write out;

  std::cout << "******************************************************************" << std::endl;
  std::cout << "*** 2D reverse time migration using finite differences           *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                                   *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017       *" << std::endl;
  std::cout << "******************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Read data header
//  par.read_headers();
  
  // Build geomery from input parameters
  par.build_geometry();

  // Read velocity
  par.read_velocity();
 
  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Obtain data geometry and apertures from input parameters
//  par.data_geometry();
  
  // Read and verbose source
  par.read_src(1);
  
  // Verbose some useful parameters
  par.verbose_par();

  // Initialize tapering for receiver propagation
  taper_init(par);

  // Initialize some variables for the main calculation
  par.ntSnap=0;

  std::cout << "*********** MAIN CALCULATION BEGINS! ********** " << std::endl;
  // Initialize parameters for forward and backward propagations
  fwd.prop_init(par);
  bwd.prop_init(par);

  // Create velocity frames
  par.vel_frames(fwd.Vshot,par.rbc);
  par.vel_frames(bwd.Vshot,par.rbc);

  // Initialize parameters for RTM
  rtm.rtm_init(par,fwd,bwd);

  // Perform Reverse-Time Migration in each shot gather
  rtm.rtm(par);

  // Construct the images
  rtm.build_images(par);

  std::cout << "*********** MAIN CALCULATION FINISHES! **********" << std::endl;

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  std::cout << "Write the output for RTM" << std::endl;  
//  if(par.eic == 0) {
    out.write_image(par,rtm.ICorr.data(),"rtm_cic");
//  }
//  else {
//    out.write_image(par,rtm.IEIC.data(),"rtm_eic");
//  }

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " 
            << std::chrono::duration<double>(t2-t1).count() 
            << "secs." << std::endl; 
  std::cout << "rtm2d: Finishing successfully!!!" << std::endl;  

  hclose();
  return 0;
}
