/*
  **************************************************************
  *** Main program for 2D stacking using MPI.                ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2018 ***
  **************************************************************
 */ 
#include <mpi.h>
#include <stdio.h>
#include "param_mpi.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <vector>
#include "allocv.h"
#include <cmath>
#include "stack.h"
#include <iostream>
#include <iomanip>
#include <chrono>
#include <string>
#include <string.h>

int main(int argc, char *argv[]) {
  // Start SEP parameters
  initpar(argc, argv);

  // Define class member objects
  ParamMPI par;

 /**********************************************************************
 * BEGINS MPI INITIALIZATION TASKS
 ************************************************************************/
  // Initialize MPI
  int ierr = MPI_Init(NULL, NULL);

  // Get the process ID and the number of processors
  int this_process, nprocess;
  ierr = MPI_Comm_rank(MPI_COMM_WORLD, &this_process);
  ierr = MPI_Comm_size(MPI_COMM_WORLD, &nprocess);

  // Report number of processes
  if(this_process == 0) std::cout << "No. of processors: " << nprocess << std::endl;

  // Define some variables
  int verbose, nfiles;
  char in_pref[100];

  // Read history and parameters in processor 0
  if(this_process == 0) {
    std::cout << "Reading input parameters" << std::endl;
    if(0 == getch("verbose","d",&verbose)) verbose=0;
    if(0 == getch("in_pref","s",&in_pref)) { 
      seperr("You need to specify the prefix common to input files!");
    }

    // Count the number of files for stacking
    nfiles=par.read_num_files(in_pref);

    // Read history file: It gets information from one of the input files
    std::cout << "Read history file" << std::endl;
    par.read_history();
  }

  // Barrier
  if(0 != MPI_Barrier(MPI_COMM_WORLD)) seperr("Error in barrier(1)");

  MPI_Status process_status;

  if(nprocess > 1) { // Tasks to execute if using MPI

    /*** Broadcast variables to processes ***/
    ierr = MPI_Bcast(&nfiles,1,MPI_INT,0,MPI_COMM_WORLD);
    if(ierr != 0) seperr("Cannot broadcast nfiles");

    int len;
    if(this_process == 0) len=strlen(in_pref);
    ierr = MPI_Bcast(&len,1,MPI_INT,0,MPI_COMM_WORLD);
    ierr = MPI_Bcast(&in_pref,len+1,MPI_CHAR,0,MPI_COMM_WORLD);
    if(ierr != 0) seperr("Cannot broadcast in_pref");

    ierr = MPI_Bcast(&verbose,1,MPI_INT,0,MPI_COMM_WORLD);
    if(ierr != 0) seperr("Cannot broadcast verbose");

    ierr = MPI_Bcast(&nfiles,1,MPI_INT,0,MPI_COMM_WORLD);
    if(ierr != 0) seperr("Cannot broadcast nfiles");

    ierr = MPI_Bcast(&par.n,1,MPI_INT,0,MPI_COMM_WORLD);
    if(ierr != 0) seperr("Cannot broadcast n");

    ierr = MPI_Bcast(&par.nsize,1,MPI_INT,0,MPI_COMM_WORLD);
    if(ierr != 0) seperr("Cannot broadcast nsize");
  }
  else { // No using MPI
    std::cout << "stack3d_mpi will run sequencially" << std::endl;
    this_process=0;
    nprocess=1;
  }

  // Barrier
  if(0 != MPI_Barrier(MPI_COMM_WORLD)) seperr("Error in barrier(2)");

 /***********************************************************************
 * FINISHES MPI INITIALIZATION TASKS
 ************************************************************************/

  // Allocate working vector
  par.allocate_vector();

  // Define starting and ending files
  int startfile=nfiles*this_process/nprocess+1;
  int endfile=  nfiles*(this_process+1)/nprocess;

  if(this_process == 0) { 
    std::cout << "************************************************************" << std::endl;
    std::cout << "*** 2D stacking using MPI.                                 *" << std::endl;
    std::cout << "*** Intended to speed up stacking of extended volumes      *" << std::endl;
    std::cout << "*** Author: Alejandro Cabrales                             *" << std::endl;
    std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2018 *" << std::endl;
    std::cout << "************************************************************" << std::endl;

    // Verbose parameters in processor 0
    std::cout << "\nFILES AND MPI PARAMETERS: " << std::endl;
    std::cout << "nprocess: " << nprocess << std::endl;
    std::cout << "in_pref: " << in_pref << std::endl;
    std::cout << "nfiles: " << nfiles << std::endl;
  }

  // Verbose which files each processor is going to handle
  std::cout << this_process << ": startfile: " << startfile << std::endl;
  std::cout << this_process << ": endfile: " << endfile << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Define a string vector containing the names of input files
  std::string pref1=std::string(in_pref)+"_00";
  std::string pref2=std::string(in_pref)+"_0";
  std::string pref3=std::string(in_pref)+"_";

  // Barrier
  if(0 != MPI_Barrier(MPI_COMM_WORLD)) seperr("Error in barrier(3)");

  std::string stack_file=std::string(in_pref)+"_stack.H";
  if(this_process == 0) {
    par.write_history(stack_file);
  }

  // Loop around the offsets (hx and hy combined)
//  for(int jext=0; jext<par.nhxy; jext++) {
//    if(this_process == 0) std::cout << "\nPass No. " << jext+1 
//                              << " of " << par.nhxy << std::endl; 

    // Initialize stacking function
    Stack stack;
    stack.stack_init(par.n);

    // Perform partial stacking in individual processors
    std::string file_in;
    for(int jfile=startfile; jfile<=endfile; jfile++) {
      long long nfile=jfile;
      if(jfile >=1 && jfile < 10) {
        file_in=pref1+std::to_string(nfile)+".H";
      }
      else if(jfile >=10 && jfile < 100) {
        file_in=pref2+std::to_string(nfile)+".H";
      }
      else if(jfile >= 100 && jfile < 1000) {
        file_in=pref3+std::to_string(nfile)+".H";
      }
      if(verbose==1) std::cout << this_process << ": Read file: " << file_in << std::endl;
      par.read_binary(file_in.c_str());
      stack.stack(par.d);
    }
  
    if(this_process !=0) {
      // Secondary processors send to main processor
      MPI_Send(stack.S.data(),par.n,MPI_FLOAT,0,1,MPI_COMM_WORLD);
    }
    else {
      // Main processor receives from secondary processors and performs final stack
      for(int jprocess=1; jprocess<nprocess; jprocess++) {
        MPI_Recv(par.d.data(),par.n,MPI_FLOAT,jprocess,1,MPI_COMM_WORLD,&process_status);
        stack.stack(par.d);
      }
    }
  
    if(this_process == 0) {
      if(verbose==1) std::cout << "Save stack binary in processor 0" << std::endl; 
      par.write_binary(stack.S,stack_file);
    }
//  }

  // Barrier
  if(0 != MPI_Barrier(MPI_COMM_WORLD)) seperr("Error in barrier(4)");

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  if(this_process == 0) {
    std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count()
                                  << "secs." << std::endl; 
    std::cout << "stack3d_mpi: Finishing successfully!!!" << std::endl;  
  }

  // Finalize MPI
  if(0 != MPI_Finalize()) seperr("Problem finalizing MPI in stack3d_mpi!");
}
