#include "read.h"
#include "wave_prop.h"
#include "alloc.h"
#include "allocv.h"
#include <iostream>

// Initialize the forward propagation
void WaveProp::prop_init(Read par_in) {
  // Shorcuts for frequently object's members invokation
  par=par_in;
  nkgx=par.nkgx;
  nkz =par.nkz;
  nkgxEx=nkgx+2*buff;
  nkzEx =nkz +2*buff;

  // Coefficients of 2nd deriv., 10th order approx.
  // Courtesy of Ettore Biondi
  c0 = -2.927222222222182   ; 
  c1 =  1.666666666666642   ;
  c2 = -0.238095238095235   ;
  c3 =  0.039682539682539   ;
  c4 = -0.004960317460317   ;
  c5 =  3.174603174603100e-4;

  // Create coefficients with the spacings 
  float idx=1.0/par.dx;
  float idz=1.0/par.dz;
  float idt2=par.dt*par.dt;
  alloc(ax,buff);
  alloc(az,buff);
  ax[0]=c0*idx*idx*idt2;
  ax[1]=c1*idx*idx*idt2;
  ax[2]=c2*idx*idx*idt2;
  ax[3]=c3*idx*idx*idt2;
  ax[4]=c4*idx*idx*idt2;
  ax[5]=c5*idx*idx*idt2;
                
  az[0]=c0*idz*idz*idt2;
  az[1]=c1*idz*idz*idt2;
  az[2]=c2*idz*idz*idt2;
  az[3]=c3*idz*idz*idt2;
  az[4]=c4*idz*idz*idt2;
  az[5]=c5*idz*idz*idt2;

  // Allocate wavefield vectors
  int n=nkgx*nkz;
  allocV(p,  n);
  allocV(pm1,n);
  allocV(pm2,n);
  allocV(dpcur,nkgxEx*nkzEx);  

  // Allocate velocity arrays
  allocV(Vshot,nkgx*nkz);
  allocV(Vel2,nkgx*nkz);
}

void WaveProp::init_wavefields() {
  // Initialize source wavefield arrays
  #pragma omp parallel
  {
    #pragma omp for
    for(int jgx=0; jgx<nkgx; jgx++) {
      #pragma omp simd
      for(int jz=0; jz<nkz; jz++) {
        int jj=jz+jgx*nkz;
        p[jj]     = 0.0;
        pm1[jj]   = 0.0;
      }
    }
    // Take velocity square from velocity frame
    #pragma omp for
    for(int jgx=0; jgx<nkgx; jgx++) {
      #pragma omp simd
      for(int jz=0; jz<nkz; jz++) {
        int jj=jz+jgx*nkz;
        Vel2[jj]=Vshot[jj]*Vshot[jj]; 
      }
    }
  }
}

void WaveProp::prop(int adj) {
  // Update the wavefield frames
  pm2.swap(pm1);
  pm1.swap(p);

  if(adj==1) {
    // Multiply pcur by Vel2 before Laplacian computation
    #pragma omp parallel
    {
      #pragma omp for
      for(int jgx=buff; jgx<nkgx+buff; jgx++) {
        #pragma omp simd simdlen(8)
        for(int jz=buff; jz<nkz+buff; jz++) { 
          int jj=jz+jgx*nkzEx;
          int ii=jz-buff+(jgx-buff)*nkz;
          dpcur[jj]=pm1[ii]*Vel2[ii];
        }
      }
  
      // Compute the Laplacian
      #pragma omp for schedule(guided)
      for(int jgx=buff; jgx<nkgx+buff; jgx++) {
        #pragma omp simd simdlen(8)
        for(int jz=buff; jz<nkz+buff; jz++) {
          int jj=jz+jgx*nkzEx;
          int ii=jz-buff+(jgx-buff)*nkz;
          p[ii] = 2.*pm1[ii]-pm2[ii]+
            (az[0]+ax[0]) *dpcur[jj]+
            az[1]*(dpcur[jj+1]      +dpcur[jj-1])+
            az[2]*(dpcur[jj+2]      +dpcur[jj-2])+
            az[3]*(dpcur[jj+3]      +dpcur[jj-3])+
            az[4]*(dpcur[jj+4]      +dpcur[jj-4])+
            az[5]*(dpcur[jj+5]      +dpcur[jj-5])+
            ax[1]*(dpcur[jj+1*nkzEx]+dpcur[jj-1*nkzEx])+
            ax[2]*(dpcur[jj+2*nkzEx]+dpcur[jj-2*nkzEx])+
            ax[3]*(dpcur[jj+3*nkzEx]+dpcur[jj-3*nkzEx])+
            ax[4]*(dpcur[jj+4*nkzEx]+dpcur[jj-4*nkzEx])+
            ax[5]*(dpcur[jj+5*nkzEx]+dpcur[jj-5*nkzEx]);
        }
      }
    }
  }
  else {
    #pragma omp parallel
    {
      #pragma omp for
      for(int jgx=buff; jgx<nkgx+buff; jgx++) {
        #pragma omp simd simdlen(8)
        for(int jz=buff; jz<nkz+buff; jz++) {
          int jj=jz+jgx*nkzEx;
          int ii=jz-buff+(jgx-buff)*nkz;
          dpcur[jj]=pm1[ii];
        }
      }
      // Compute the second derivative
      #pragma omp for schedule(guided)
      for(int jgx=buff; jgx<nkgx+buff; jgx++) {
        #pragma omp simd simdlen(8)
        for(int jz=buff; jz<nkz+buff; jz++) {
          int jj=jz+jgx*nkzEx;
          int ii=jz-buff+(jgx-buff)*nkz;
          p[ii] = 2.*pm1[ii]-pm2[ii]+ Vel2[ii]*( 
            (ax[0]+az[0]) *dpcur[jj]+
            az[1]*(dpcur[jj+1]      +dpcur[jj-1])+
            az[2]*(dpcur[jj+2]      +dpcur[jj-2])+
            az[3]*(dpcur[jj+3]      +dpcur[jj-3])+
            az[4]*(dpcur[jj+4]      +dpcur[jj-4])+
            az[5]*(dpcur[jj+5]      +dpcur[jj-5])+
            ax[1]*(dpcur[jj+1*nkzEx]+dpcur[jj-1*nkzEx])+
            ax[2]*(dpcur[jj+2*nkzEx]+dpcur[jj-2*nkzEx])+
            ax[3]*(dpcur[jj+3*nkzEx]+dpcur[jj-3*nkzEx])+
            ax[4]*(dpcur[jj+4*nkzEx]+dpcur[jj-4*nkzEx])+
            ax[5]*(dpcur[jj+5*nkzEx]+dpcur[jj-5*nkzEx]));
        }
      }
    }
  }
}

