#include "read.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "taper.h"
#include "alloc.h"
#include "allocv.h"
#include <vector>
#include <cmath>
#include "wemva.h"
#include "snap.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include <iostream>
#include <string>
#include <chrono>
#include <numeric>

void Wemva::wemva_init(Read &par,WaveProp &fwd_in,
                       WaveProp &bwd_in,WaveProp &scat_in) {
  // Shorcuts for frequently object's members invokation
  pad=par.pad;
  zpad=par.zpad;
  nkgx=par.nkgx;
  nkz=par.nkz;
  nx=par.nx;
  nz=par.nz;
  nxE=par.nxE;
  nzE=par.nzE;
  nt=par.nt;
  napL=par.napL;
  napR=par.napR;
  dt=par.dt;
  dx=par.dx;
  dz=par.dz;
  use_taper=par.use_taper;
  first_cdp=par.first_cdp;
  last_cdp=par.last_cdp;
  nh=par.nh;
  hmax=par.hmax;
  datumS=par.datumS;
  datumR=par.datumR;
  fwd=&fwd_in; 
  bwd=&bwd_in; 
  scat=&scat_in; 
  dipole=par.dipole;

  int n1, n2;
  n1=nkgx*nkz*(2*nh+1);
  n2=nxE*nzE*(2*nh+1);

  // Allocate vectors
  allocV(pert,n1);
  allocV(IPert,n2);
  allocV(pshot,n1);

  // Create reflectivity frames 
  par.refl_frames(pshot);

  // Vertical position of the source (dipole source)
  jsrc1=datumS  +pad+zpad+(pad+napL+par.shot_pos)*nkz;
  if(dipole) jsrc2=datumS+1+pad+zpad+(pad+napL+par.shot_pos)*nkz;

  // Reading data (traces from one shot gather)
  par.read_shot(1);
}

void Wemva::wemva(Read &par) {
  // Initialize/update wavefield arrays
  (*fwd).init_wavefields();
  (*scat).init_wavefields();
  (*bwd).init_wavefields();

  // Temporary auxiliary vectors
  std::vector<float> paux1(nkgx*nkz,0.);
  std::vector<float> paux2(nkgx*nkz,0.);
  std::vector<float> paux3(nkgx*nkz,0.);

  // Initialize scalar
  scaleS.scale_init(par,(*fwd).Vel2,1);

   /*
   WEMVA is performed in three stages, explained in SEP report 168
   (Cabrales-Vargas et al., 2017). Seven wavefield propagations are
   required (not five!) using random boundary conditions (Clapp, 2009) 
   */
  std::cout << "Perform WEMVA" << std::endl;

   // 1st STAGE: Forward propagation of source wavefield
  for(int jt=0; jt<nt+2; jt++) {
    // Forward propagation of source wavefield
    if(jt>1) (*fwd).prop(0); // 1st propagation

    if(jt>0 && jt<=nt) {
      // Prepare the dipole source
      paux1[jsrc1]= par.src[jt-1];
      if(dipole) paux1[jsrc2]=-par.src[jt-1];

      // Scale, then inject the source
      scaleS.scale(0,1,paux1,(*fwd).p);
    }

    // Save snapshots: source wavefield
//    if(jt<nt) snapshot(par,jt,fwd.p,1,"snap1");
  }

  // Swap arrays
  (*fwd).p.swap((*fwd).pm1);

  // Initialize scalar
  scaleR.scale_init(par,(*fwd).Vel2,0);
  scaleScat.scale_init(par,(*fwd).Vel2,1);

  // Initialize scatter receiver
  scatterR.scatter_init(par,paux2);

  // Initialize source imaging condition
  scatterS.scatter_init(par,(*fwd).p);

  #pragma omp parallel for
  for(int jj=0; jj<paux1.size(); jj++) {
    paux1[jj]=0.;
    paux2[jj]=0.;
    paux3[jj]=0.;
  }

  // 2nd STAGE: Backward propagation of receiver wavefield + scattering
  // and xcorr with back-propagated source wavefield
  for(int jt=nt+1; jt>=-2; jt--) {
    // Backward propagation of source wavefield
    if(jt>=0) (*fwd).prop(0); // 3rd propagation

    if(jt>0 && jt<nt+1) {
      // Prepare the dipole source wavelet
      paux3[jsrc1]= par.src[jt-1];
      if(dipole) paux3[jsrc2]=-par.src[jt-1];

      // Scale, then inject the source
      scaleS.scale(0,1,paux3,(*fwd).p);
    }

    // Backward propagation of receiver wavefield
    if(jt<nt-0) (*bwd).prop(1); // 2nd propagation

    if(jt>0 && jt<nt+1) {
      // Prepare the dipole data, and inject
//      #pragma omp parallel for
      for(int jgx=0; jgx<nkgx; jgx++) {
        int j1=datumR  +pad+zpad+jgx*nkz;
        int j2=datumR+1+pad+zpad+jgx*nkz;
        int j3=jt-1+jgx*nt;
        (*bwd).p[j1]+=par.rcv[j3];
        if(dipole) (*bwd).p[j2]-=par.rcv[j3];
      }
    }

    // Scale the rec. wavefield after propagation
    scaleR.scale(1,0,paux2,(*bwd).pm1);

    if(jt<nt-1) {
      // Propagate the scattered receiver wavefield
      (*scat).prop(1); // 4th propagation

      // Use damping function in receiver propagation wavefields
      if(use_taper==1) { 
        taper((*scat).pm2,(*scat).pm1,(*scat).p);
      }
    }

    if(remainder(jt,par.scatStep)==0) {
      if(jt>=-1 && jt<nt-0) {
        // Scatter the receiver wavefield upon reflectivity (ext. domain)
        scatterR.scatter(0,1,pshot,(*scat).p); 
      }
        
      if(jt<nt-0) {
        // Scale the scattered receiver wavefield after propagation
        scaleScat.scale(1,0,paux1,(*scat).p);
      }

      // Save snapshots: receiver scattered wavefield
//      if(jt>=0) snapshot(par,jt,scat.p,0,"snap2");
    }

    // Use damping function in source propagation wavefields
//    if(use_taper==1) {
//      taper((*fwd).pm2,(*fwd).pm1,(*fwd).p);
//    }

    if(remainder(jt,par.scatStep)==0 && jt>0 && jt<nt-0) {
      // Apply imaging condition (ext. domain): S*deltaR
      scatterS.scatter(1,1,pert,paux1); 
    }
  }

  // Prepare the receiver wavefield for forward-propagatiom
  (*bwd).p.swap((*bwd).pm1);

  // Re-initialize source wavefield arrays
  (*fwd).init_wavefields();
  (*scat).init_wavefields();
 
  // Initialize scalar
  scaleS.scale_init(par,(*fwd).Vel2,1);
  scaleScat.scale_init(par,(*fwd).Vel2,1);

  // Initialize scatter source
  scatterS.scatter_init(par,(*fwd).pm1);

  // Initialize receiver imaging condition
  scatterR.scatter_init(par,paux2);

  #pragma omp parallel for
  for(int jj=0; jj<paux1.size(); jj++) {
    paux1[jj]=0.;
    paux2[jj]=0.;
    paux3[jj]=0.;
  }

  // 3rd STAGE: Forward propagation of source wavefield + scattering
  // and xcorr with forward-propagated receiver wavefield
  for(int jt=-2; jt<nt; jt++) {
    // Forward propagation of receiver wavefield
    (*bwd).prop(1); // 5th propagation

    if(jt>0) {
      // Prepare the dipole data
//      #pragma omp parallel for
      for(int jgx=0; jgx<nkgx; jgx++) {
        int j1=datumR  +pad+zpad+jgx*nkz;
        int j2=datumR+1+pad+zpad+jgx*nkz;
        int j3=jt-1+jgx*nt;
        (*bwd).p[j1]+=par.rcv[j3];
        if(dipole) (*bwd).p[j2]-=par.rcv[j3];
      }
    }

    // Scale the rec. wavefield after propagation
    scaleR.scale(1,0,paux2,(*bwd).p);

    // Forward propagation of source wavefield
    if(jt>1) (*fwd).prop(0); // 6th propagation

    if(jt>0) {
      // Prepare the dipole source
      paux3[jsrc1]= par.src[jt-1];
      if(dipole) paux3[jsrc2]=-par.src[jt-1];

      // Scale, then inject the source
      scaleS.scale(0,1,paux3,(*fwd).p);
    }

    // Save snapshots: receiver wavefield
//    snapshot(par,jt,fwd.p,0,"snap3");

    if(jt>2) {
      // Propagate the scattered source wavefield
      (*scat).prop(0);
    }

    if(jt>1 && remainder(jt,par.scatStep)==0) {
      // Scatter the source wavefield upon reflectivity (ext. domain)
      scatterS.scatter(0,0,pshot,paux1); 

      // Scale the scattered source wavefield before propagation
      scaleScat.scale(0,1,paux1,(*scat).p);

      // Damping function applied to the scattered wavefield
      if(use_taper == 1) {
        taper((*scat).pm2,(*scat).pm1,(*scat).p);
      }
    }

    // Use damping function in receiver propagation wavefields
//    if(use_taper==1) {
//      taper((*fwd).pm2,(*fwd).pm1,(*fwd).p);
//    }

    // Apply imaging condition (ext. domain): R*deltaS
    if(remainder(jt,par.scatStep)==0 && jt>0) {
      scatterR.scatter(1,1,pert,(*scat).p); 
    }
  }
}

void Wemva::build_pert(Read par) {
  for(int jh=0; jh<2*nh+1; jh++){
//  #pragma omp parallel for
    for(int jgx=0; jgx<nkgx; jgx++){
      int jx=floor(par.sgx[jgx]/dx)+pad+napL;
      if(jx>=0 && jx<=nxE) {
//        #pragma omp simd
        for(int jz=0; jz<nkz; jz++) {
//          int j1=jz+jh*nkz+jgx*(2*nh+1)*nkz;
//          int j2=jz+jh*nzE+jx*(2*nh+1)*nzE;
          int j1=jz+jgx*nkz+jh*nkgx*nkz;
          int j2=jz+jx*nzE+jh*nxE*nzE;
          IPert[j2]=pert[j1];
        }
      }
    }
  }
}
