/*
  **************************************************************
  *** Main program for 2D WEMVA modeling.                    ***
  *** WEMVA using finite differences                         ***
  *** Author: Alejandro Cabrales                             ***
  *** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  **************************************************************
 */ 
#include "read.h"
#include "write.h"
#include "taper.h"
#include "seplib.h"
#include "sep3d.h"
#include "sep_pars_external.h"
#include "interp1d.h"
#include "grad.h"
#include "wave_prop.h"
#include "scale.h"
#include "scatter.h"
#include "alloc.h"
#include <cmath>
#include "wemva.h"
#include "dso.h"
#include "wemva_op.h"
#include <sep.main>
#include <iostream>
#include <iomanip>
#include <chrono>

int MAIN() {
  // Define class member objects
  Read par;
  WaveProp fwd, scat, bwd;
  Wemva wemva,wemva_adj;
  Write out;

  std::cout << "*************************************************************" << std::endl;
  std::cout << "*** 2D WEMVA modeling using FD  method                      *" << std::endl;
  std::cout << "*** Author: Alejandro Cabrales                              *" << std::endl;
  std::cout << "*** Stanford Exploration Project/Petroleos Mexicanos, 2017  *" << std::endl;
  std::cout << "*************************************************************" << std::endl;

  // Initializing wall time
  auto t1=std::chrono::system_clock::now();

  // Read the inout history file
  par.read_parameters();

  // Read data history
  par.read_history();
  
  // Build geomery from input parameters
  par.build_geometry();

  // Read velocity
  par.read_velocity();
  
  // Read perturbation (either in slowness or reflectivity)
  par.read_pert(par.adj);
  
  // Determine time sampling according to Courant-Fredrich-Lewis condition
  par.courant();
  
  // Read and verbose source
  par.read_src(1);
  
  // Verbose some useful parameters
  par.verbose_par();
 
  // Initialize tapering for receiver propagation
  taper_init(par);

  // Initialize some variables for the main loop
  par.ntSnap=0;

  std::cout << "*********** MAIN COMPUTATION BEGINS! ********** " << std::endl;
  // Initialize parameters for forward and backward propagations
  fwd.prop_init(par);
  bwd.prop_init(par);
  scat.prop_init(par);

  // Create velocity frames
  par.vel_frames(fwd.Vshot,par.rbc);
  par.vel_frames(bwd.Vshot,par.rbc);
  par.vel_frames(scat.Vshot,par.rbc);

  // Initialize WEMVA operator
  Read &ppar=par;
  Wemva &pwemva=wemva;
  Wemva &pwemva_adj=wemva_adj;

  // This objects are for B-spline smoother, adapted from Tang's thesis
//  Axis axd, axk, ayd, ayk, azd, azk;
  int use_dso;
//  if(par.dso == -1) {
//    use_dso=0;
//  }
//  else {
    use_dso=par.dso;
//  }
  Dso dso(par,use_dso);
  Wemva_op op(ppar,pwemva,pwemva_adj,fwd,bwd,scat,dso,1);

  // Initialize model and data vectors
  int nmod=par.nkgx*par.nkz*(2*par.nh+1);
  int ndat=par.nkgx*par.nkz*(2*par.nh+1);
  std::vector<float> model(nmod,0.);
  std::vector<float> data(ndat,0.);

  if(par.adj==1) {
  // Adjoint WEMVA
    data.swap(wemva.pshot);
    op.op(1,0,model,data);
    wemva.pert.swap(model);
  }
  else {
  // Forward WEMVA
    model.swap(wemva.pshot);
    op.op(0,0,model,data);
    wemva.pert.swap(data);
  } 

  // Build perturbation (on reflectivity or slowness squared)
  wemva.build_pert(par);

  std::cout << "*********** MAIN COMPUTATION FINISHES! ********** " << std::endl;

  // Initialize the output
  out.write_init(par);
  
  // Write the output
  if(par.adj == 0) {
    std::cout << "Write the output for WEMVA: Pert. in the image" << std::endl;  
    out.write_image(par,wemva.IPert.data(),"wemva");
  }
  else {
    std::cout << "Write the output for WEMVA: Pert. in the background" << std::endl;  
    out.write_pert(par,wemva.IPert.data(),"wemva");
  }

  // Finalizing wall time
  auto t2=std::chrono::system_clock::now();
  std::cout << "Elapsed time: " << std::chrono::duration<double>(t2-t1).count() << "secs." << std::endl; 
  std::cout << "wemva2d: Finishing successfully!!!" << std::endl;  

  hclose();

 return 0; 
}
